# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import threading
import time
import unittest
from collections.abc import Sequence
from concurrent.futures import (  # pylint: disable=no-name-in-module
    ThreadPoolExecutor,
)
from logging import WARNING, getLogger
from platform import system
from typing import Any
from unittest import TestCase
from unittest.mock import Mock, patch

import grpc
from google.protobuf.duration_pb2 import (  # pylint: disable=no-name-in-module
    Duration,
)
from google.rpc.error_details_pb2 import (  # pylint: disable=no-name-in-module
    RetryInfo,
)
from grpc import ChannelCredentials, Compression, StatusCode, server

from opentelemetry.exporter.otlp.proto.common.trace_encoder import (
    encode_spans,
)
from opentelemetry.exporter.otlp.proto.grpc.exporter import (  # noqa: F401
    _RETRYABLE_ERROR_CODES,
    InvalidCompressionValueException,
    OTLPExporterMixin,
    environ_to_compression,
)
from opentelemetry.exporter.otlp.proto.grpc.version import __version__
from opentelemetry.proto.collector.trace.v1.trace_service_pb2 import (
    ExportTraceServiceRequest,
    ExportTraceServiceResponse,
)
from opentelemetry.proto.collector.trace.v1.trace_service_pb2_grpc import (
    TraceServiceServicer,
    TraceServiceStub,
    add_TraceServiceServicer_to_server,
)
from opentelemetry.sdk.environment_variables import (
    _OTEL_PYTHON_EXPORTER_OTLP_GRPC_CREDENTIAL_PROVIDER,
    OTEL_EXPORTER_OTLP_COMPRESSION,
    OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED,
)
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import InMemoryMetricReader
from opentelemetry.sdk.trace import ReadableSpan, _Span
from opentelemetry.sdk.trace.export import (
    SpanExporter,
    SpanExportResult,
)
from opentelemetry.semconv._incubating.attributes.otel_attributes import (
    OtelComponentTypeValues,
)
from opentelemetry.test.mock_test_classes import IterEntryPoint

logger = getLogger(__name__)


# The below tests use this test SpanExporter and Spans, but are testing the
# underlying behavior in the mixin. A MetricExporter or LogRecordExporter could
# just as easily be used.
class OTLPSpanExporterForTesting(
    SpanExporter,
    OTLPExporterMixin[
        ReadableSpan,
        ExportTraceServiceRequest,
        SpanExportResult,
        TraceServiceStub,
    ],
):
    def __init__(self, **kwargs):
        super().__init__(
            TraceServiceStub,
            SpanExportResult,
            component_type=OtelComponentTypeValues.OTLP_GRPC_SPAN_EXPORTER,
            signal="traces",
            meter_provider=kwargs.pop("meter_provider", None),
            **kwargs,
        )

    def _translate_data(
        self, data: Sequence[ReadableSpan]
    ) -> ExportTraceServiceRequest:
        return encode_spans(data)

    def _count_data(self, data: Sequence[ReadableSpan]) -> int:
        return len(data)

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        return self._export(spans)

    @property
    def _exporting(self):
        return "traces"

    def shutdown(self, timeout_millis: float = 30_000, **kwargs):
        return OTLPExporterMixin.shutdown(self, timeout_millis, **kwargs)


class TraceServiceServicerWithExportParams(TraceServiceServicer):
    def __init__(
        self,
        export_result: StatusCode,
        optional_retry_nanos: int | None = None,
        optional_export_sleep: float | None = None,
        optional_error_details: str | None = None,
    ):
        self.export_result = export_result
        self.optional_export_sleep = optional_export_sleep
        self.optional_retry_nanos = optional_retry_nanos
        self.num_requests = 0
        self.optional_error_details = optional_error_details

    # pylint: disable=invalid-name,unused-argument
    def Export(self, request, context):
        self.num_requests += 1
        if self.optional_export_sleep:
            time.sleep(self.optional_export_sleep)
        if self.export_result != StatusCode.OK and self.optional_retry_nanos:
            context.set_trailing_metadata(
                (
                    (
                        "google.rpc.retryinfo-bin",
                        RetryInfo(
                            retry_delay=Duration(
                                nanos=self.optional_retry_nanos
                            )
                        ).SerializeToString(),
                    ),
                )
            )
        context.set_code(self.export_result)
        if self.optional_error_details:
            context.set_details(self.optional_error_details)

        return ExportTraceServiceResponse()


class ThreadWithReturnValue(threading.Thread):
    def __init__(
        self,
        target=None,
        args=(),
    ):
        super().__init__(target=target, args=args)
        self._return = None

    def run(self):
        try:
            if self._target is not None:  # type: ignore
                self._return = self._target(*self._args, **self._kwargs)  # type: ignore
        finally:
            # Avoid a refcycle if the thread is running a function with
            # an argument that has a member that points to the thread.
            del self._target, self._args, self._kwargs  # type: ignore

    def join(self, timeout: float | None = None) -> Any:
        super().join(timeout=timeout)
        return self._return


# pylint: disable-next=too-many-public-methods
class TestOTLPExporterMixin(TestCase):
    def setUp(self):
        self.server = server(ThreadPoolExecutor(max_workers=10))

        self.server.add_insecure_port("127.0.0.1:4317")

        self.server.start()

        self.metric_reader = InMemoryMetricReader()
        self.meter_provider = MeterProvider(
            metric_readers=[self.metric_reader]
        )
        self.exporter = OTLPSpanExporterForTesting(
            insecure=True, meter_provider=self.meter_provider
        )
        self.span = _Span(
            "a",
            context=Mock(
                **{
                    "trace_state": {"a": "b", "c": "d"},
                    "span_id": 10217189687419569865,
                    "trace_id": 67545097771067222548457157018666467027,
                }
            ),
        )

    def tearDown(self):
        self.server.stop(None)

    @patch("opentelemetry.exporter.otlp.proto.grpc.exporter.insecure_channel")
    @patch("opentelemetry.exporter.otlp.proto.grpc.exporter.secure_channel")
    def test_otlp_exporter_endpoint(self, mock_secure, mock_insecure):
        expected_endpoint = "localhost:4317"
        endpoints = [
            (
                "http://localhost:4317",
                None,
                mock_insecure,
            ),
            (
                "localhost:4317",
                None,
                mock_secure,
            ),
            (
                "http://localhost:4317",
                True,
                mock_insecure,
            ),
            (
                "localhost:4317",
                True,
                mock_insecure,
            ),
            (
                "http://localhost:4317",
                False,
                mock_secure,
            ),
            (
                "localhost:4317",
                False,
                mock_secure,
            ),
            (
                "https://localhost:4317",
                False,
                mock_secure,
            ),
            (
                "https://localhost:4317",
                None,
                mock_secure,
            ),
            (
                "https://localhost:4317",
                True,
                mock_secure,
            ),
        ]
        for endpoint, insecure, mock_method in endpoints:
            OTLPSpanExporterForTesting(endpoint=endpoint, insecure=insecure)
            self.assertEqual(
                1,
                mock_method.call_count,
                f"expected {mock_method} to be called for {endpoint} {insecure}",
            )
            self.assertEqual(
                expected_endpoint,
                mock_method.call_args[0][0],
                f"expected {expected_endpoint} got {mock_method.call_args[0][0]} {endpoint}",
            )
            mock_method.reset_mock()

    def test_environ_to_compression(self):
        with patch.dict(
            "os.environ",
            {
                "test_gzip": "gzip",
                "test_gzip_caseinsensitive_with_whitespace": " GzIp ",
                "test_invalid": "some invalid compression",
            },
        ):
            self.assertEqual(
                environ_to_compression("test_gzip"), Compression.Gzip
            )
            self.assertEqual(
                environ_to_compression(
                    "test_gzip_caseinsensitive_with_whitespace"
                ),
                Compression.Gzip,
            )
            self.assertIsNone(
                environ_to_compression("missing_key"),
            )
            with self.assertRaises(InvalidCompressionValueException):
                environ_to_compression("test_invalid")

    # pylint: disable=no-self-use
    @patch("opentelemetry.exporter.otlp.proto.grpc.exporter.insecure_channel")
    @patch.dict("os.environ", {})
    def test_otlp_exporter_otlp_compression_unspecified(
        self, mock_insecure_channel
    ):
        """No env or kwarg should be NoCompression"""
        OTLPSpanExporterForTesting(insecure=True)
        mock_insecure_channel.assert_called_once_with(
            "localhost:4317",
            compression=Compression.NoCompression,
            options=(
                (
                    "grpc.primary_user_agent",
                    "OTel-OTLP-Exporter-Python/" + __version__,
                ),
            ),
        )

    @patch.dict(
        "os.environ",
        {
            _OTEL_PYTHON_EXPORTER_OTLP_GRPC_CREDENTIAL_PROVIDER: "credential_provider"
        },
    )
    @patch("opentelemetry.exporter.otlp.proto.grpc.exporter.entry_points")
    def test_that_credential_gets_passed_to_exporter(self, mock_entry_points):
        credential = ChannelCredentials(None)

        def f():
            return credential

        mock_entry_points.configure_mock(
            return_value=[IterEntryPoint("custom_credential", f)]
        )
        exporter = OTLPSpanExporterForTesting(insecure=False)
        # pylint: disable=protected-access
        assert exporter._credentials is credential

    @patch.dict(
        "os.environ",
        {
            _OTEL_PYTHON_EXPORTER_OTLP_GRPC_CREDENTIAL_PROVIDER: "credential_provider"
        },
    )
    def test_that_missing_entry_point_raises_exception(self):
        with self.assertRaises(RuntimeError):
            OTLPSpanExporterForTesting(insecure=False)

    @patch.dict(
        "os.environ",
        {
            _OTEL_PYTHON_EXPORTER_OTLP_GRPC_CREDENTIAL_PROVIDER: "credential_provider"
        },
    )
    @patch("opentelemetry.exporter.otlp.proto.grpc.exporter.entry_points")
    def test_that_entry_point_returning_bad_type_raises_exception(
        self, mock_entry_points
    ):
        def f():
            return 1

        mock_entry_points.configure_mock(
            return_value=[IterEntryPoint("custom_credential", f)]
        )
        with self.assertRaises(RuntimeError):
            OTLPSpanExporterForTesting(insecure=False)

    # pylint: disable=no-self-use, disable=unused-argument
    @patch(
        "opentelemetry.exporter.otlp.proto.grpc.exporter.ssl_channel_credentials"
    )
    @patch("opentelemetry.exporter.otlp.proto.grpc.exporter.secure_channel")
    @patch.dict("os.environ", {})
    def test_no_credentials_ssl_channel_called(
        self, secure_channel, mock_ssl_channel
    ):
        OTLPSpanExporterForTesting(insecure=False)
        self.assertTrue(mock_ssl_channel.called)

    # pylint: disable=no-self-use
    @patch("opentelemetry.exporter.otlp.proto.grpc.exporter.insecure_channel")
    @patch.dict("os.environ", {OTEL_EXPORTER_OTLP_COMPRESSION: "gzip"})
    def test_otlp_exporter_otlp_compression_envvar(
        self, mock_insecure_channel
    ):
        """Just OTEL_EXPORTER_OTLP_COMPRESSION should work"""
        OTLPSpanExporterForTesting(insecure=True)
        mock_insecure_channel.assert_called_once_with(
            "localhost:4317",
            compression=Compression.Gzip,
            options=(
                (
                    "grpc.primary_user_agent",
                    "OTel-OTLP-Exporter-Python/" + __version__,
                ),
            ),
        )

    @patch.dict(
        "os.environ", {OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED: " true "}
    )
    def test_shutdown(self):
        add_TraceServiceServicer_to_server(
            TraceServiceServicerWithExportParams(StatusCode.OK),
            self.server,
        )
        exporter = OTLPSpanExporterForTesting(
            insecure=True, meter_provider=self.meter_provider
        )
        self.assertEqual(
            exporter.export([self.span]), SpanExportResult.SUCCESS
        )
        metrics_data = self.metric_reader.get_metrics_data()
        scope_metrics = metrics_data.resource_metrics[0].scope_metrics[0]
        self.assertEqual(scope_metrics.scope.name, "opentelemetry-sdk")
        metrics = sorted(scope_metrics.metrics, key=lambda m: m.name)
        self.assertEqual(len(metrics), 3)
        self.assertEqual(
            metrics[0].name, "otel.sdk.exporter.operation.duration"
        )
        self.assert_standard_metric_attrs(
            metrics[0].data.data_points[0].attributes
        )
        self.assertEqual(metrics[1].name, "otel.sdk.exporter.span.exported")
        self.assert_standard_metric_attrs(
            metrics[1].data.data_points[0].attributes
        )
        self.assertEqual(metrics[2].name, "otel.sdk.exporter.span.inflight")
        self.assert_standard_metric_attrs(
            metrics[2].data.data_points[0].attributes
        )

        exporter.shutdown()
        with self.assertLogs(level=WARNING) as warning:
            self.assertEqual(
                exporter.export([self.span]), SpanExportResult.FAILURE
            )
            self.assertEqual(
                warning.records[0].message,
                "Exporter already shutdown, ignoring batch",
            )

    @unittest.skipIf(
        system() == "Windows",
        "For gRPC + windows there's some added delay in the RPCs which breaks the assertion over amount of time passed.",
    )
    def test_shutdown_interrupts_export_retry_backoff(self):
        add_TraceServiceServicer_to_server(
            TraceServiceServicerWithExportParams(
                StatusCode.UNAVAILABLE,
            ),
            self.server,
        )

        export_thread = ThreadWithReturnValue(
            target=self.exporter.export, args=([self.span],)
        )
        with self.assertLogs(level=WARNING) as warning:
            begin_wait = time.time()
            export_thread.start()
            # Wait a bit for export to fail and the backoff sleep to start
            time.sleep(0.05)
            # The code should now be in a 1 second backoff.
            # pylint: disable=protected-access
            self.assertFalse(self.exporter._shutdown_in_progress.is_set())
            self.exporter.shutdown()
            self.assertTrue(self.exporter._shutdown_in_progress.is_set())
            export_result = export_thread.join()
            end_wait = time.time()
            self.assertEqual(export_result, SpanExportResult.FAILURE)
            # Shutdown should have interrupted the sleep.
            self.assertTrue(end_wait - begin_wait < 0.2)
            self.assertEqual(
                warning.records[1].message,
                "Shutdown in progress, aborting retry.",
            )

    def test_export_over_closed_grpc_channel(self):
        # pylint: disable=protected-access

        add_TraceServiceServicer_to_server(
            TraceServiceServicerWithExportParams(StatusCode.OK),
            self.server,
        )
        self.exporter.export([self.span])
        self.exporter.shutdown()
        data = self.exporter._translate_data([self.span])
        with self.assertRaises(ValueError) as err:
            self.exporter._client.Export(request=data)
        self.assertEqual(
            str(err.exception), "Cannot invoke RPC on closed channel!"
        )

    @unittest.skipIf(
        system() == "Windows",
        "For gRPC + windows there's some added delay in the RPCs which breaks the assertion over amount of time passed.",
    )
    @patch.dict(
        "os.environ", {OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED: "true"}
    )
    def test_retry_info_is_respected(self):
        mock_trace_service = TraceServiceServicerWithExportParams(
            StatusCode.UNAVAILABLE,
            optional_retry_nanos=200000000,  # .2 seconds
        )
        add_TraceServiceServicer_to_server(
            mock_trace_service,
            self.server,
        )
        exporter = OTLPSpanExporterForTesting(
            insecure=True, timeout=10, meter_provider=self.meter_provider
        )
        before = time.time()
        self.assertEqual(
            exporter.export([self.span]),
            SpanExportResult.FAILURE,
        )
        after = time.time()
        self.assertEqual(mock_trace_service.num_requests, 6)
        # 1 second plus wiggle room so the test passes consistently.
        self.assertAlmostEqual(after - before, 1, 1)

        metrics_data = self.metric_reader.get_metrics_data()
        scope_metrics = metrics_data.resource_metrics[0].scope_metrics[0]
        self.assertEqual(scope_metrics.scope.name, "opentelemetry-sdk")
        metrics = sorted(scope_metrics.metrics, key=lambda m: m.name)
        self.assertEqual(len(metrics), 3)
        self.assertEqual(
            metrics[0].name, "otel.sdk.exporter.operation.duration"
        )
        self.assert_standard_metric_attrs(
            metrics[0].data.data_points[0].attributes
        )
        self.assertEqual(
            metrics[0].data.data_points[0].attributes["error.type"],
            "_InactiveRpcError",
        )
        self.assertEqual(
            metrics[0]
            .data.data_points[0]
            .attributes["rpc.response.status_code"],
            "UNAVAILABLE",
        )
        self.assertEqual(metrics[1].name, "otel.sdk.exporter.span.exported")
        self.assert_standard_metric_attrs(
            metrics[1].data.data_points[0].attributes
        )
        self.assertEqual(
            metrics[1].data.data_points[0].attributes["error.type"],
            "_InactiveRpcError",
        )
        self.assertNotIn(
            "rpc.response.status_code",
            metrics[1].data.data_points[0].attributes,
        )
        self.assertEqual(metrics[2].name, "otel.sdk.exporter.span.inflight")
        self.assert_standard_metric_attrs(
            metrics[2].data.data_points[0].attributes
        )
        self.assertNotIn(
            "error.type", metrics[2].data.data_points[0].attributes
        )
        self.assertNotIn(
            "rpc.response.status_code",
            metrics[2].data.data_points[0].attributes,
        )

    @unittest.skipIf(
        system() == "Windows",
        "For gRPC + windows there's some added delay in the RPCs which breaks the assertion over amount of time passed.",
    )
    def test_retry_not_made_if_would_exceed_timeout(self):
        mock_trace_service = TraceServiceServicerWithExportParams(
            StatusCode.UNAVAILABLE
        )
        add_TraceServiceServicer_to_server(
            mock_trace_service,
            self.server,
        )
        exporter = OTLPSpanExporterForTesting(insecure=True, timeout=4)
        before = time.time()
        self.assertEqual(
            exporter.export([self.span]),
            SpanExportResult.FAILURE,
        )
        after = time.time()
        # Our retry starts with a 1 second backoff then doubles.
        # First call at time 0, second at time 1, third at time 3, fourth would exceed timeout.
        self.assertEqual(mock_trace_service.num_requests, 3)
        # There's a +/-20% jitter on each backoff.
        self.assertTrue(2.35 < after - before < 3.65)

    @unittest.skipIf(
        system() == "Windows",
        "For gRPC + windows there's some added delay in the RPCs which breaks the assertion over amount of time passed.",
    )
    def test_timeout_set_correctly(self):
        mock_trace_service = TraceServiceServicerWithExportParams(
            StatusCode.UNAVAILABLE, optional_export_sleep=0.25
        )
        add_TraceServiceServicer_to_server(
            mock_trace_service,
            self.server,
        )
        exporter = OTLPSpanExporterForTesting(insecure=True, timeout=1.4)
        # Should timeout after 1.4 seconds. First attempt takes .25 seconds
        # Then a 1 second sleep, then deadline exceeded after .15 seconds,
        # mid way through second call.
        with self.assertLogs(level=WARNING) as warning:
            before = time.time()
            # Eliminate the jitter.
            with patch("random.uniform", return_value=1):
                self.assertEqual(
                    exporter.export([self.span]),
                    SpanExportResult.FAILURE,
                )
            after = time.time()
            self.assertEqual(
                "Failed to export traces to localhost:4317, error code: StatusCode.DEADLINE_EXCEEDED, error details: Deadline Exceeded",
                warning.records[-1].message,
            )
            self.assertEqual(mock_trace_service.num_requests, 2)
            self.assertAlmostEqual(after - before, 1.4, 1)

    def test_channel_options_set_correctly(self):
        """Test that gRPC channel options are set correctly for keepalive and reconnection"""
        # This test verifies that the channel is created with the right options
        # We patch grpc.insecure_channel to ensure it is called without errors
        with patch(
            "opentelemetry.exporter.otlp.proto.grpc.exporter.insecure_channel"
        ) as mock_channel:
            OTLPSpanExporterForTesting(insecure=True)
            self.assertTrue(mock_channel.called)

    def test_otlp_headers_from_env(self):
        # pylint: disable=protected-access
        # This ensures that there is no other header than standard user-agent.
        self.assertEqual(
            self.exporter._headers,
            (),
        )

    @patch.dict(
        "os.environ", {OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED: "true"}
    )
    def test_permanent_failure(self):
        exporter = OTLPSpanExporterForTesting(
            insecure=True, meter_provider=self.meter_provider
        )
        with self.assertLogs(level=WARNING) as warning:
            add_TraceServiceServicer_to_server(
                TraceServiceServicerWithExportParams(
                    StatusCode.ALREADY_EXISTS,
                    optional_error_details="This already exists.",
                ),
                self.server,
            )
            self.assertEqual(
                exporter.export([self.span]), SpanExportResult.FAILURE
            )
            self.assertEqual(
                warning.records[-1].message,
                "Failed to export traces to localhost:4317, error code: StatusCode.ALREADY_EXISTS, error details: This already exists.",
            )

        metrics_data = self.metric_reader.get_metrics_data()
        scope_metrics = metrics_data.resource_metrics[0].scope_metrics[0]
        self.assertEqual(scope_metrics.scope.name, "opentelemetry-sdk")
        metrics = sorted(scope_metrics.metrics, key=lambda m: m.name)
        self.assertEqual(len(metrics), 3)
        self.assertEqual(
            metrics[0].name, "otel.sdk.exporter.operation.duration"
        )
        self.assert_standard_metric_attrs(
            metrics[0].data.data_points[0].attributes
        )
        self.assertEqual(
            metrics[0].data.data_points[0].attributes["error.type"],
            "_InactiveRpcError",
        )
        self.assertEqual(
            metrics[0]
            .data.data_points[0]
            .attributes["rpc.response.status_code"],
            "ALREADY_EXISTS",
        )
        self.assertEqual(metrics[1].name, "otel.sdk.exporter.span.exported")
        self.assert_standard_metric_attrs(
            metrics[1].data.data_points[0].attributes
        )
        self.assertEqual(
            metrics[1].data.data_points[0].attributes["error.type"],
            "_InactiveRpcError",
        )
        self.assertNotIn(
            "rpc.response.status_code",
            metrics[1].data.data_points[0].attributes,
        )
        self.assertEqual(metrics[2].name, "otel.sdk.exporter.span.inflight")
        self.assert_standard_metric_attrs(
            metrics[2].data.data_points[0].attributes
        )
        self.assertNotIn(
            "error.type", metrics[2].data.data_points[0].attributes
        )
        self.assertNotIn(
            "rpc.response.status_code",
            metrics[2].data.data_points[0].attributes,
        )

    def test_unavailable_reconnects(self):
        """Test that the exporter reconnects on UNAVAILABLE error"""
        add_TraceServiceServicer_to_server(
            TraceServiceServicerWithExportParams(StatusCode.UNAVAILABLE),
            self.server,
        )

        # Spy on grpc.insecure_channel to verify it's called for reconnection
        with patch(
            "opentelemetry.exporter.otlp.proto.grpc.exporter.insecure_channel",
            side_effect=grpc.insecure_channel,
        ) as mock_insecure_channel:
            # Mock sleep to avoid waiting
            with patch("time.sleep"):
                # We expect FAILURE because the server keeps returning UNAVAILABLE
                # but we want to verify reconnection attempts happened
                self.exporter.export([self.span])

        # Verify that we attempted to reinitialize the channel (called insecure_channel)
        # Since the initial channel was created in setUp (unpatched), this call
        # must be from the reconnection logic.
        self.assertTrue(mock_insecure_channel.called)

    def test_retryable_error_codes_initialization(self):
        # pylint: disable=protected-access
        self.assertEqual(
            self.exporter._retryable_error_codes, _RETRYABLE_ERROR_CODES
        )
        custom_codes = [StatusCode.INTERNAL, StatusCode.UNKNOWN]
        exporter = OTLPSpanExporterForTesting(
            insecure=True, retryable_error_codes=custom_codes
        )
        self.assertEqual(
            exporter._retryable_error_codes, frozenset(custom_codes)
        )

    @patch.dict(
        "os.environ",
        {
            "OTEL_PYTHON_EXPORTER_OTLP_GRPC_RETRYABLE_ERROR_CODES": ",INTERNAL, unknown,,,dEAdline_Exceeded "
        },
    )
    def test_retryable_error_codes_initialization_from_env(self):
        expected_codes = frozenset(
            {
                StatusCode.INTERNAL,
                StatusCode.UNKNOWN,
                StatusCode.DEADLINE_EXCEEDED,
            }
        )
        exporter = OTLPSpanExporterForTesting()
        # pylint: disable=protected-access
        self.assertEqual(exporter._retryable_error_codes, expected_codes)

    @unittest.skipIf(
        system() == "Windows",
        "For gRPC + windows there's some added delay in the RPCs which breaks the assertion over amount of time passed.",
    )
    def test_retryable_error_codes_custom(self):
        # Test that a custom error code is retried if specified
        custom_codes = [StatusCode.INTERNAL]
        mock_trace_service = TraceServiceServicerWithExportParams(
            StatusCode.INTERNAL,
            optional_retry_nanos=200000000,  # .2 seconds
        )
        add_TraceServiceServicer_to_server(
            mock_trace_service,
            self.server,
        )
        exporter = OTLPSpanExporterForTesting(
            insecure=True, retryable_error_codes=custom_codes, timeout=10
        )

        self.assertEqual(
            exporter.export([self.span]),
            SpanExportResult.FAILURE,
        )

        self.assertEqual(mock_trace_service.num_requests, 6)

        # Test that a default retryable code is NOT retried if not in custom_codes
        mock_trace_service.num_requests = 0
        mock_trace_service.export_result = StatusCode.UNAVAILABLE
        self.assertEqual(
            exporter.export([self.span]),
            SpanExportResult.FAILURE,
        )
        self.assertEqual(mock_trace_service.num_requests, 1)

    def assert_standard_metric_attrs(self, attributes):
        self.assertEqual(
            attributes["otel.component.type"], "otlp_grpc_span_exporter"
        )
        self.assertTrue(
            attributes["otel.component.name"].startswith(
                "otlp_grpc_span_exporter/"
            )
        )
        self.assertEqual(attributes["server.address"], "localhost")
        self.assertEqual(attributes["server.port"], 4317)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=unexpected-keyword-arg

import gzip
import threading
import unittest
import zlib
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timezone
from email.utils import format_datetime
from unittest.mock import Mock, patch

# pylint: disable-next=import-error
from opentelemetry.exporter.http.transport._base import (
    BaseHTTPResult,
    BaseHTTPTransport,
)

# pylint: disable-next=import-error
from opentelemetry.exporter.otlp.common._http import (
    Compression,
    OTLPHTTPClient,
    _extract_retry_after,
)


@contextmanager
def _mock_clock(
    shutdown_event: Mock | None = None,
) -> Iterator[Callable[[float], None]]:
    _now = [0.0]

    def advance(delta: float) -> None:
        _now[0] += delta

    def get_time() -> float:
        return _now[0]

    if shutdown_event is not None:

        def _wait(duration: float) -> bool:
            advance(duration)
            return False

        shutdown_event.wait.side_effect = _wait

    with patch(
        "opentelemetry.exporter.otlp.common._http.time.time",
        side_effect=get_time,
    ):
        yield advance


@dataclass(frozen=True, slots=True)
class _TestHTTPResult(BaseHTTPResult):
    response_headers: dict = field(default_factory=dict)

    # pylint: disable-next=no-self-use
    def content(self) -> bytes:
        return b""

    def headers(self):
        return self.response_headers


class _TestHTTPTransport(BaseHTTPTransport):
    def __init__(self, *results, connection_errors=()):
        self.results = list(results)
        self.requests = []
        self.closed = False
        self._connection_errors = set(connection_errors)

    def request(
        self,
        method,
        url,
        *,
        headers=None,
        timeout=None,
        data=None,
    ):
        self.requests.append(
            {
                "method": method,
                "url": url,
                "headers": headers,
                "timeout": timeout,
                "data": data,
            }
        )
        result = self.results.pop(0)
        if callable(result):
            result = result()
        if isinstance(result, Exception):
            raise result
        return result

    def is_connection_error(self, exception):
        return exception in self._connection_errors

    def close(self):
        self.closed = True


class TestOTLPHTTPClient(unittest.TestCase):
    @staticmethod
    def _client(
        transport,
        *,
        timeout=5.0,
        compression=Compression.NONE,
        jitter=0.0,
    ):
        return OTLPHTTPClient(
            transport=transport,
            endpoint="http://example.test/v1/traces",
            timeout=timeout,
            compression=compression,
            headers={"content-type": "application/x-protobuf"},
            kind="spans",
            jitter=jitter,
        )

    def test_export_success_status_codes(self):
        cases = (
            (200, "OK"),
            (204, "No Content"),
            (302, "Found"),
        )

        for status_code, reason in cases:
            with self.subTest(status_code=status_code):
                transport = _TestHTTPTransport(
                    _TestHTTPResult(status_code=status_code, reason=reason)
                )
                client = self._client(transport)

                result = client.export(b"payload")

                self.assertTrue(result.success)
                self.assertEqual(result.status_code, status_code)
                self.assertEqual(result.reason, reason)
                self.assertIsNone(result.error)

    @patch(
        "opentelemetry.exporter.otlp.common._http.time.time",
        side_effect=(100.0, 100.0, 100.0),
    )
    def test_export_request_arguments(self, mock_time):
        transport = _TestHTTPTransport(
            _TestHTTPResult(status_code=200, reason="OK")
        )
        client = self._client(transport, timeout=3.0)

        client.export(b"payload")

        self.assertEqual(len(transport.requests), 1)
        self.assertEqual(
            transport.requests[0],
            {
                "method": "POST",
                "url": "http://example.test/v1/traces",
                "headers": {"content-type": "application/x-protobuf"},
                "timeout": 3.0,
                "data": b"payload",
            },
        )
        self.assertEqual(mock_time.call_count, 3)

    def test_export_compresses_payload(self):
        cases = (
            (Compression.NONE, lambda data: data, None),
            (Compression.GZIP, gzip.decompress, "gzip"),
            (Compression.DEFLATE, zlib.decompress, "deflate"),
        )

        for compression, decompress, expected_encoding in cases:
            with self.subTest(compression=compression):
                transport = _TestHTTPTransport(
                    _TestHTTPResult(status_code=200, reason="OK")
                )
                client = self._client(transport, compression=compression)

                result = client.export(b"payload")

                self.assertTrue(result.success)
                self.assertEqual(
                    decompress(transport.requests[0]["data"]), b"payload"
                )
                headers = transport.requests[0]["headers"]
                if expected_encoding is None:
                    self.assertNotIn("Content-Encoding", headers)
                else:
                    self.assertEqual(
                        headers["Content-Encoding"], expected_encoding
                    )

    def test_export_retryable_status_codes(self):
        cases = (
            (429, "Too Many Requests"),
            (502, "Bad Gateway"),
            (503, "Service Unavailable"),
            (504, "Gateway Timeout"),
        )

        for status_code, reason in cases:
            with self.subTest(status_code=status_code):
                shutdown_event = Mock(spec=threading.Event)
                shutdown_event.is_set.return_value = False
                shutdown_event.wait.return_value = False
                transport = _TestHTTPTransport(
                    _TestHTTPResult(
                        status_code=status_code,
                        reason=reason,
                    ),
                    _TestHTTPResult(status_code=200, reason="OK"),
                )
                client = self._client(transport)
                # pylint: disable-next=protected-access
                client._shutdown_event = shutdown_event

                result = client.export(b"payload")

                self.assertTrue(result.success)
                self.assertEqual(len(transport.requests), 2)
                shutdown_event.wait.assert_called_once_with(1.0)

    def test_export_connection_errors(self):
        error = RuntimeError("connection failed")
        transport = _TestHTTPTransport(
            _TestHTTPResult(error=error),
            _TestHTTPResult(status_code=200, reason="OK"),
            connection_errors={error},
        )
        client = self._client(transport)

        result = client.export(b"payload")

        self.assertTrue(result.success)
        self.assertEqual(len(transport.requests), 2)
        self.assertAlmostEqual(transport.requests[0]["timeout"], 5.0, 2)
        self.assertLessEqual(
            transport.requests[1]["timeout"],
            transport.requests[0]["timeout"],
        )
        self.assertGreater(transport.requests[1]["timeout"], 0.0)

    def test_export_non_retryable_errors(self):
        exception = RuntimeError("request failed")
        cases = (
            (
                _TestHTTPResult(status_code=400, reason="Bad Request"),
                400,
                "Bad Request",
                None,
            ),
            (
                _TestHTTPResult(status_code=408, reason="Request Timeout"),
                408,
                "Request Timeout",
                None,
            ),
            (
                _TestHTTPResult(
                    status_code=500, reason="Internal Server Error"
                ),
                500,
                "Internal Server Error",
                None,
            ),
            (
                _TestHTTPResult(error=exception),
                None,
                None,
                exception,
            ),
            (
                exception,
                None,
                None,
                exception,
            ),
        )

        for (
            response,
            expected_status_code,
            expected_reason,
            expected_error,
        ) in cases:
            with self.subTest(response=type(response).__name__):
                transport = _TestHTTPTransport(response)
                client = self._client(transport)

                result = client.export(b"payload")

                self.assertFalse(result.success)
                self.assertEqual(result.status_code, expected_status_code)
                self.assertEqual(result.reason, expected_reason)
                self.assertIs(result.error, expected_error)

    def test_export_with_shutdown(self):
        shutdown_event = Mock(spec=threading.Event)
        shutdown_event.is_set.return_value = True
        transport = _TestHTTPTransport(
            _TestHTTPResult(status_code=503, reason="Service Unavailable")
        )
        client = self._client(transport)
        # pylint: disable-next=protected-access
        client._shutdown_event = shutdown_event

        result = client.export(b"payload")

        self.assertFalse(result.success)
        self.assertEqual(result.status_code, 503)
        self.assertEqual(result.reason, "Service Unavailable")
        shutdown_event.wait.assert_not_called()

    def test_shutdown_closes_transport(self):
        transport = _TestHTTPTransport()
        client = self._client(transport)

        client.shutdown()

        self.assertTrue(transport.closed)

    def test_export_timeout_decreases_per_retry(self):
        shutdown_event = Mock(spec=threading.Event)
        shutdown_event.is_set.return_value = False
        transport = _TestHTTPTransport(
            _TestHTTPResult(status_code=503, reason="Service Unavailable"),
            _TestHTTPResult(status_code=503, reason="Service Unavailable"),
            _TestHTTPResult(status_code=200, reason="OK"),
        )
        client = self._client(transport, timeout=10.0, jitter=0.0)
        # pylint: disable-next=protected-access
        client._shutdown_event = shutdown_event

        with _mock_clock(shutdown_event):
            result = client.export(b"payload")

        # retry=0: wait(1.0) -> time=1.0, retry=1: wait(2.0) -> time=3.0, success
        self.assertTrue(result.success)
        self.assertAlmostEqual(transport.requests[0]["timeout"], 10.0)
        self.assertAlmostEqual(transport.requests[1]["timeout"], 9.0)
        self.assertAlmostEqual(transport.requests[2]["timeout"], 7.0)

    def test_export_backoff_exhausts_remaining_timeout(self):
        shutdown_event = Mock(spec=threading.Event)
        shutdown_event.is_set.return_value = False
        transport = _TestHTTPTransport(
            _TestHTTPResult(status_code=503, reason="Service Unavailable"),
            _TestHTTPResult(status_code=503, reason="Service Unavailable"),
        )
        # timeout=1.5: retry=0 backoff=1.0 fits -> wait(1.0) -> time=1.0
        # retry=1 backoff=2.0 > 0.5 remaining -> give up
        client = self._client(transport, timeout=1.5, jitter=0.0)
        # pylint: disable-next=protected-access
        client._shutdown_event = shutdown_event

        with _mock_clock(shutdown_event):
            result = client.export(b"payload")

        self.assertFalse(result.success)
        self.assertEqual(result.status_code, 503)
        self.assertEqual(len(transport.requests), 2)
        shutdown_event.wait.assert_called_once_with(1.0)

    def test_export_exhausts_max_retries(self):
        shutdown_event = Mock(spec=threading.Event)
        shutdown_event.is_set.return_value = False
        transport = _TestHTTPTransport(
            *[_TestHTTPResult(status_code=503, reason="Service Unavailable")]
            * 6
        )
        client = self._client(transport, timeout=1000.0, jitter=0.0)
        # pylint: disable-next=protected-access
        client._shutdown_event = shutdown_event

        with _mock_clock(shutdown_event):
            result = client.export(b"payload")

        self.assertFalse(result.success)
        self.assertEqual(len(transport.requests), 6)
        self.assertEqual(shutdown_event.wait.call_count, 5)
        self.assertEqual(
            [call.args[0] for call in shutdown_event.wait.call_args_list],
            [1.0, 2.0, 4.0, 8.0, 16.0],
        )

    def test_export_connection_error_gets_reduced_timeout(self):
        stale_error = RuntimeError("stale connection")
        transport = _TestHTTPTransport(
            _TestHTTPResult(status_code=200, reason="OK"),
            connection_errors={stale_error},
        )

        with _mock_clock() as advance:

            def _slow_connection_error() -> _TestHTTPResult:
                advance(2.0)
                return _TestHTTPResult(error=stale_error)

            transport.results.insert(0, _slow_connection_error)
            client = self._client(transport, timeout=5.0)
            result = client.export(b"payload")

        # _submit: deadline=0+5=5.0, after first request time=2.0, remaining=3.0
        self.assertTrue(result.success)
        self.assertEqual(len(transport.requests), 2)
        self.assertAlmostEqual(transport.requests[1]["timeout"], 3.0)

    def test_export_retry_after_header_used_as_backoff(self):
        shutdown_event = Mock(spec=threading.Event)
        shutdown_event.is_set.return_value = False
        shutdown_event.wait.return_value = False
        transport = _TestHTTPTransport(
            _TestHTTPResult(
                status_code=429,
                reason="Too Many Requests",
                response_headers={"retry-after": "5"},
            ),
            _TestHTTPResult(status_code=200, reason="OK"),
        )
        client = self._client(transport, timeout=60.0)
        # pylint: disable-next=protected-access
        client._shutdown_event = shutdown_event

        result = client.export(b"payload")

        self.assertTrue(result.success)
        self.assertEqual(len(transport.requests), 2)
        shutdown_event.wait.assert_called_once_with(5.0)

    def test_export_retry_after_header_exhausts_timeout(self):
        shutdown_event = Mock(spec=threading.Event)
        shutdown_event.is_set.return_value = False
        transport = _TestHTTPTransport(
            _TestHTTPResult(
                status_code=503,
                reason="Service Unavailable",
                response_headers={"retry-after": "10"},
            ),
        )
        client = self._client(transport, timeout=3.0, jitter=0.0)
        # pylint: disable-next=protected-access
        client._shutdown_event = shutdown_event

        with _mock_clock(shutdown_event):
            result = client.export(b"payload")

        self.assertFalse(result.success)
        self.assertEqual(result.status_code, 503)
        self.assertEqual(len(transport.requests), 1)
        shutdown_event.wait.assert_not_called()

    def test_export_retry_after_http_date(self):
        base = 1_700_000_000.0
        shutdown_event = Mock(spec=threading.Event)
        shutdown_event.is_set.return_value = False
        shutdown_event.wait.return_value = False
        retry_at = format_datetime(
            datetime.fromtimestamp(base + 30, timezone.utc), usegmt=True
        )
        transport = _TestHTTPTransport(
            _TestHTTPResult(
                status_code=503,
                reason="Service Unavailable",
                response_headers={"retry-after": retry_at},
            ),
            _TestHTTPResult(status_code=200, reason="OK"),
        )
        client = self._client(transport, timeout=120.0)
        # pylint: disable-next=protected-access
        client._shutdown_event = shutdown_event

        with patch(
            "opentelemetry.exporter.otlp.common._http.time.time",
            return_value=base,
        ):
            result = client.export(b"payload")

        self.assertTrue(result.success)
        self.assertEqual(len(transport.requests), 2)
        shutdown_event.wait.assert_called_once_with(30.0)

    def test_export_retry_after_http_date_in_past(self):
        base = 1_700_000_000.0
        shutdown_event = Mock(spec=threading.Event)
        shutdown_event.is_set.return_value = False
        shutdown_event.wait.return_value = False
        retry_at = format_datetime(
            datetime.fromtimestamp(base - 30, timezone.utc), usegmt=True
        )
        transport = _TestHTTPTransport(
            _TestHTTPResult(
                status_code=429,
                reason="Too Many Requests",
                response_headers={"retry-after": retry_at},
            ),
            _TestHTTPResult(status_code=200, reason="OK"),
        )
        client = self._client(transport, timeout=120.0)
        # pylint: disable-next=protected-access
        client._shutdown_event = shutdown_event

        with patch(
            "opentelemetry.exporter.otlp.common._http.time.time",
            return_value=base,
        ):
            result = client.export(b"payload")

        self.assertTrue(result.success)
        self.assertEqual(len(transport.requests), 2)
        shutdown_event.wait.assert_called_once_with(0.0)

    def test_export_retry_after_malformed(self):
        shutdown_event = Mock(spec=threading.Event)
        shutdown_event.is_set.return_value = False
        shutdown_event.wait.return_value = False
        transport = _TestHTTPTransport(
            _TestHTTPResult(
                status_code=503,
                reason="Service Unavailable",
                response_headers={"retry-after": "not-a-date"},
            ),
            _TestHTTPResult(status_code=200, reason="OK"),
        )
        client = self._client(transport, timeout=60.0, jitter=0.0)
        # pylint: disable-next=protected-access
        client._shutdown_event = shutdown_event

        result = client.export(b"payload")

        self.assertTrue(result.success)
        self.assertEqual(len(transport.requests), 2)
        # No usable Retry-After -> exponential backoff: 2**0 * uniform(1, 1).
        shutdown_event.wait.assert_called_once_with(1.0)

    def test_extract_retry_after_edge_cases(self):
        cases = (
            ("5", 5.0),
            ("  7  ", 7.0),
            ("-3", 0.0),
            ("nan", None),
            ("inf", None),
            ("garbage", None),
        )
        for value, expected in cases:
            with self.subTest(value=value):
                self.assertEqual(
                    _extract_retry_after(
                        _TestHTTPResult(
                            response_headers={"retry-after": value}
                        )
                    ),
                    expected,
                )

        with self.subTest(value="absent"):
            self.assertIsNone(_extract_retry_after(_TestHTTPResult()))



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import math
from collections.abc import Iterator

import pytest
from grpc import Compression as GRPCCompression
from inline_snapshot import snapshot

from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter as GRPCSpanExporter,
)
from opentelemetry.exporter.otlp.proto.http import (
    Compression as HTTPCompression,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as HTTPSpanExporter,
)
from opentelemetry.sdk.trace import Tracer, TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor, SpanExporter
from opentelemetry.test._otlp_test_server import OtlpProtoTestServer
from opentelemetry.trace import Link, SpanContext, StatusCode, TraceFlags

from . import CUSTOM_HEADERS, ExporterConfig, _attrs_to_dict

TRACE_EXPORTER_CONFIGS: list[ExporterConfig[SpanExporter]] = [
    ExporterConfig(
        id="http",
        exporter_class=HTTPSpanExporter,
        kwargs={"endpoint": "http://localhost:4318/v1/traces"},
    ),
    ExporterConfig(
        id="http-deflate",
        exporter_class=HTTPSpanExporter,
        kwargs={
            "endpoint": "http://localhost:4318/v1/traces",
            "compression": HTTPCompression.Deflate,
        },
    ),
    ExporterConfig(
        id="http-gzip",
        exporter_class=HTTPSpanExporter,
        kwargs={
            "endpoint": "http://localhost:4318/v1/traces",
            "compression": HTTPCompression.Gzip,
        },
    ),
    ExporterConfig(
        id="http-headers",
        exporter_class=HTTPSpanExporter,
        kwargs={
            "endpoint": "http://localhost:4318/v1/traces",
            "headers": CUSTOM_HEADERS,
        },
    ),
    ExporterConfig(
        id="grpc",
        exporter_class=GRPCSpanExporter,
        kwargs={"insecure": True},
    ),
    ExporterConfig(
        id="grpc-gzip",
        exporter_class=GRPCSpanExporter,
        kwargs={"insecure": True, "compression": GRPCCompression.Gzip},
    ),
    ExporterConfig(
        id="grpc-headers",
        exporter_class=GRPCSpanExporter,
        kwargs={"insecure": True, "headers": CUSTOM_HEADERS},
    ),
]


class TestTracesExporter:
    @pytest.fixture(
        scope="class", params=TRACE_EXPORTER_CONFIGS, ids=lambda c: c.id
    )
    def config(self, request) -> ExporterConfig[SpanExporter]:
        return request.param

    @pytest.fixture(scope="class")
    def tracer(
        self,
        config: ExporterConfig[SpanExporter],
        server: OtlpProtoTestServer,
    ) -> Iterator[Tracer]:
        tracer_provider = TracerProvider()
        tracer_provider.add_span_processor(SimpleSpanProcessor(config.build()))
        try:
            yield tracer_provider.get_tracer(__name__)
        finally:
            tracer_provider.shutdown()

    @pytest.fixture(autouse=True)
    def clear_server(self, server: OtlpProtoTestServer) -> None:
        server.clear()

    def test_simple_span_name(
        self, tracer: Tracer, server: OtlpProtoTestServer
    ):
        with tracer.start_as_current_span("my-span"):
            pass

        recorded = server.get_span(timeout=5.0)
        assert recorded.span.name == "my-span"

    def test_span_attributes(
        self, tracer: Tracer, server: OtlpProtoTestServer
    ):
        with tracer.start_as_current_span(
            "attrs-span",
            attributes={
                "str_key": "hello",
                "int_key": 42,
                "float_key": 3.14,
                "bool_key": True,
            },
        ):
            pass

        recorded = server.get_span(timeout=5.0)
        attrs = _attrs_to_dict(recorded.span.attributes)
        assert math.isclose(attrs.pop("float_key"), 3.14, abs_tol=1e-5)
        assert attrs == snapshot(
            {"str_key": "hello", "int_key": 42, "bool_key": True}
        )

    def test_nested_spans_parent_child(
        self, tracer: Tracer, server: OtlpProtoTestServer
    ):
        with tracer.start_as_current_span("foo"):
            with tracer.start_as_current_span("bar"):
                with tracer.start_as_current_span("baz"):
                    pass

        spans = {
            r.span.name: r.span
            for r in server.get_spans(count=3, timeout=10.0)
        }
        assert set(spans.keys()) == snapshot({"bar", "baz", "foo"})
        assert spans["baz"].parent_span_id == spans["bar"].span_id
        assert spans["bar"].parent_span_id == spans["foo"].span_id
        assert spans["foo"].parent_span_id == b""

    def test_span_with_event(
        self, tracer: Tracer, server: OtlpProtoTestServer
    ):
        with tracer.start_as_current_span("event-span") as span:
            span.add_event("my-event", {"event_key": "event_val"})

        recorded = server.get_span(timeout=5.0)
        assert len(recorded.span.events) == 1
        event = recorded.span.events[0]
        assert event.name == "my-event"
        assert _attrs_to_dict(event.attributes) == snapshot(
            {"event_key": "event_val"}
        )

    def test_span_with_link(self, tracer: Tracer, server: OtlpProtoTestServer):
        link_trace_id = 0x000000000000000000000000DEADBEEF
        link_span_id = 0x00000000DEADBEF0
        link_context = SpanContext(
            trace_id=link_trace_id,
            span_id=link_span_id,
            is_remote=True,
            trace_flags=TraceFlags(0x01),
        )
        with tracer.start_as_current_span(
            "linked-span", links=[Link(link_context)]
        ):
            pass

        recorded = server.get_span(timeout=5.0)
        assert len(recorded.span.links) == 1
        link = recorded.span.links[0]
        assert link.trace_id == link_trace_id.to_bytes(16, "big")
        assert link.span_id == link_span_id.to_bytes(8, "big")

    def test_span_status_ok(self, tracer: Tracer, server: OtlpProtoTestServer):
        with tracer.start_as_current_span("ok-span") as span:
            span.set_status(StatusCode.OK)

        recorded = server.get_span(timeout=5.0)
        assert recorded.span.status.code == snapshot(1)

    def test_span_status_error(
        self, tracer: Tracer, server: OtlpProtoTestServer
    ):
        with tracer.start_as_current_span("error-span") as span:
            span.set_status(StatusCode.ERROR, "something went wrong")

        recorded = server.get_span(timeout=5.0)
        assert recorded.span.status.code == snapshot(2)
        assert recorded.span.status.message == snapshot("something went wrong")



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import dataclasses
import fnmatch
import logging
import os
from collections.abc import Callable
from typing import Any
from urllib import parse

from opentelemetry.configuration._common import load_entry_point
from opentelemetry.configuration.models import (
    AttributeNameValue,
    AttributeType,
    ExperimentalResourceDetector,
    IncludeExclude,
)
from opentelemetry.configuration.models import Resource as ResourceConfig
from opentelemetry.sdk.environment_variables import OTEL_SERVICE_NAME
from opentelemetry.sdk.resources import (
    _DEFAULT_RESOURCE,
    SERVICE_NAME,
    ProcessResourceDetector,
    Resource,
    ServiceInstanceIdResourceDetector,
    _HostResourceDetector,
)
from opentelemetry.util.types import AttributeValue

_logger = logging.getLogger(__name__)


def _coerce_bool(value: object) -> bool:
    if isinstance(value, str):
        return value.lower() not in ("false", "0", "")
    return bool(value)


def _array(coerce: Callable) -> Callable:
    return lambda value: [coerce(item) for item in value]


# Dispatch table mapping AttributeType to its coercion callable
_COERCIONS = {
    AttributeType.string: str,
    AttributeType.int: int,
    AttributeType.double: float,
    AttributeType.bool: _coerce_bool,
    AttributeType.string_array: _array(str),
    AttributeType.int_array: _array(int),
    AttributeType.double_array: _array(float),
    AttributeType.bool_array: _array(_coerce_bool),
}


def _coerce_attribute_value(attr: AttributeNameValue) -> object:
    """Coerce an attribute value to the correct Python type based on AttributeType."""
    coerce = _COERCIONS.get(attr.type)  # type: ignore[arg-type]
    return coerce(attr.value) if coerce is not None else attr.value  # type: ignore[operator]


def _parse_attributes_list(attributes_list: str) -> dict[str, str]:
    """Parse a comma-separated key=value string into a dict.

    Format is the same as OTEL_RESOURCE_ATTRIBUTES: key=value,key=value
    Values are always strings (no type coercion).
    """
    result: dict[str, str] = {}
    for item in attributes_list.split(","):
        item = item.strip()
        if not item:
            continue
        if "=" not in item:
            _logger.warning(
                "Invalid resource attribute pair in attributes_list: %s",
                item,
            )
            continue
        key, value = item.split("=", maxsplit=1)
        result[key.strip()] = parse.unquote(value.strip())
    return result


def create_resource(config: ResourceConfig | None) -> Resource:
    """Create an SDK Resource from declarative config.

    Does NOT read OTEL_RESOURCE_ATTRIBUTES. Resource detectors are only run
    when explicitly listed under detection_development.detectors in the config.
    Starts from SDK telemetry defaults (telemetry.sdk.*), merges any detected
    attributes, then merges explicit config attributes on top (highest priority).

    Args:
        config: Resource config from the parsed config file, or None.

    Returns:
        A Resource with SDK defaults, optional detector attributes, and any
        config-specified attributes merged in priority order.
    """
    # Spec requires service.name to always be present; detectors and explicit
    # config attributes can override this default.
    base = _DEFAULT_RESOURCE.merge(Resource({SERVICE_NAME: "unknown_service"}))

    if config is None:
        return base

    # attributes_list is lower priority; explicit attributes overwrite conflicts.
    config_attrs: dict[str, object] = {}
    if config.attributes_list:
        config_attrs.update(_parse_attributes_list(config.attributes_list))

    if config.attributes:
        for attr in config.attributes:
            config_attrs[attr.name] = _coerce_attribute_value(attr)

    schema_url = config.schema_url

    # Run detectors only if detection_development is configured. Collect all
    # detected attributes, apply the include/exclude filter, then merge before
    # config attributes so explicit values always win.
    result = base
    if config.detection_development:
        detected_attrs: dict[str, object] = {}
        if config.detection_development.detectors:
            for detector_config in config.detection_development.detectors:
                _run_detectors(detector_config, detected_attrs)

        filtered = _filter_attributes(
            detected_attrs, config.detection_development.attributes
        )
        if filtered:
            result = result.merge(Resource(filtered))  # type: ignore[arg-type]

    config_resource = Resource(config_attrs, schema_url)  # type: ignore[arg-type]
    return result.merge(config_resource)


def _detect_service(_config: Any) -> dict[str, AttributeValue]:
    """Service detector: generates instance ID and reads OTEL_SERVICE_NAME."""
    attrs: dict[str, AttributeValue] = dict(
        ServiceInstanceIdResourceDetector().detect().attributes
    )
    if service_name := os.environ.get(OTEL_SERVICE_NAME):
        attrs[SERVICE_NAME] = service_name
    return attrs


_RESOURCE_DETECTOR_REGISTRY: dict[
    str, Callable[[Any], dict[str, AttributeValue]]
] = {
    "service": _detect_service,
    "host": lambda _: dict(_HostResourceDetector().detect().attributes),
    "process": lambda _: dict(ProcessResourceDetector().detect().attributes),
}


def _run_detectors(
    detector_config: ExperimentalResourceDetector,
    detected_attrs: dict[str, object],
) -> None:
    """Run detectors present in a single detector config entry.

    Known detectors (service, host, process) are handled directly via
    _RESOURCE_DETECTOR_REGISTRY. All other detectors — including known
    schema fields like container that require contrib packages, and
    unknown plugin detectors captured in additional_properties — are
    loaded via the ``opentelemetry_resource_detector`` entry point group.

    The detected_attrs dict is updated in-place; later detectors overwrite
    earlier ones for the same key.
    """
    for name in dataclasses.fields(detector_config):
        value = getattr(detector_config, name.name, None)
        if value is None:
            continue
        if name.name in _RESOURCE_DETECTOR_REGISTRY:
            detected_attrs.update(
                _RESOURCE_DETECTOR_REGISTRY[name.name](value)
            )
        else:
            cls = load_entry_point(
                "opentelemetry_resource_detector", name.name
            )
            detected_attrs.update(cls(**(value or {})).detect().attributes)

    for name, plugin_config in detector_config.additional_properties.items():
        cls = load_entry_point("opentelemetry_resource_detector", name)
        detected_attrs.update(cls(**(plugin_config or {})).detect().attributes)


def _filter_attributes(
    attrs: dict[str, object], filter_config: IncludeExclude | None
) -> dict[str, object]:
    """Filter detected attribute keys using include/exclude glob patterns.

    Mirrors other SDK IncludeExcludePredicate.createPatternMatching behaviour:
    - No filter config (attributes absent) → include all detected attributes.
    - included patterns are checked first; excluded patterns are applied after.
    - An empty included list is treated as "include everything".
    """
    if filter_config is None:
        return attrs

    included = filter_config.included
    excluded = filter_config.excluded

    if not included and not excluded:
        return attrs

    result: dict[str, object] = {}
    for key, value in attrs.items():
        if included and not any(fnmatch.fnmatch(key, pat) for pat in included):
            continue
        if excluded and any(fnmatch.fnmatch(key, pat) for pat in excluded):
            continue
        result[key] = value
    return result



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from platform import system
from time import sleep
from unittest import TestCase

from pytest import mark

from opentelemetry.sdk.metrics import Counter, MeterProvider
from opentelemetry.sdk.metrics.export import (
    AggregationTemporality,
    InMemoryMetricReader,
)


class TestTimeAlign(TestCase):
    # This delay is needed for these tests to pass when they are run in
    # Windows.
    delay = 0.001

    def test_time_align_cumulative(self):
        reader = InMemoryMetricReader()
        meter_provider = MeterProvider(metric_readers=[reader])

        meter = meter_provider.get_meter("testmeter")

        counter_0 = meter.create_counter("counter_0")
        counter_1 = meter.create_counter("counter_1")

        counter_0.add(10, {"label": "value1"})
        sleep(self.delay)
        counter_0.add(10, {"label": "value2"})
        sleep(self.delay)
        counter_1.add(10, {"label": "value1"})
        sleep(self.delay)
        counter_1.add(10, {"label": "value2"})

        metrics = reader.get_metrics_data()

        data_points_0_0 = list(
            metrics.resource_metrics[0]
            .scope_metrics[0]
            .metrics[0]
            .data.data_points
        )
        data_points_0_1 = list(
            metrics.resource_metrics[0]
            .scope_metrics[0]
            .metrics[1]
            .data.data_points
        )
        self.assertEqual(len(data_points_0_0), 2)
        self.assertEqual(len(data_points_0_1), 2)

        self.assertLess(
            data_points_0_0[0].start_time_unix_nano,
            data_points_0_0[1].start_time_unix_nano,
        )
        self.assertLess(
            data_points_0_1[0].start_time_unix_nano,
            data_points_0_1[1].start_time_unix_nano,
        )
        self.assertNotEqual(
            data_points_0_0[1].start_time_unix_nano,
            data_points_0_1[0].start_time_unix_nano,
        )

        self.assertEqual(
            data_points_0_0[0].time_unix_nano,
            data_points_0_0[1].time_unix_nano,
        )
        self.assertEqual(
            data_points_0_1[0].time_unix_nano,
            data_points_0_1[1].time_unix_nano,
        )
        self.assertEqual(
            data_points_0_0[1].time_unix_nano,
            data_points_0_1[0].time_unix_nano,
        )

        counter_0.add(10, {"label": "value1"})
        sleep(self.delay)
        counter_0.add(10, {"label": "value2"})
        sleep(self.delay)
        counter_1.add(10, {"label": "value1"})
        sleep(self.delay)
        counter_1.add(10, {"label": "value2"})

        metrics = reader.get_metrics_data()

        data_points_1_0 = list(
            metrics.resource_metrics[0]
            .scope_metrics[0]
            .metrics[0]
            .data.data_points
        )
        data_points_1_1 = list(
            metrics.resource_metrics[0]
            .scope_metrics[0]
            .metrics[1]
            .data.data_points
        )

        self.assertEqual(len(data_points_1_0), 2)
        self.assertEqual(len(data_points_1_1), 2)

        self.assertLess(
            data_points_1_0[0].start_time_unix_nano,
            data_points_1_0[1].start_time_unix_nano,
        )
        self.assertLess(
            data_points_1_1[0].start_time_unix_nano,
            data_points_1_1[1].start_time_unix_nano,
        )
        self.assertNotEqual(
            data_points_1_0[1].start_time_unix_nano,
            data_points_1_1[0].start_time_unix_nano,
        )

        self.assertEqual(
            data_points_1_0[0].time_unix_nano,
            data_points_1_0[1].time_unix_nano,
        )
        self.assertEqual(
            data_points_1_1[0].time_unix_nano,
            data_points_1_1[1].time_unix_nano,
        )
        self.assertEqual(
            data_points_1_0[1].time_unix_nano,
            data_points_1_1[0].time_unix_nano,
        )

        self.assertEqual(
            data_points_0_0[0].start_time_unix_nano,
            data_points_1_0[0].start_time_unix_nano,
        )
        self.assertEqual(
            data_points_0_0[1].start_time_unix_nano,
            data_points_1_0[1].start_time_unix_nano,
        )
        self.assertEqual(
            data_points_0_1[0].start_time_unix_nano,
            data_points_1_1[0].start_time_unix_nano,
        )
        self.assertEqual(
            data_points_0_1[1].start_time_unix_nano,
            data_points_1_1[1].start_time_unix_nano,
        )

    @mark.skipif(
        system() != "Linux", reason="test failing in CI when run in Windows"
    )
    def test_time_align_delta(self):
        reader = InMemoryMetricReader(
            preferred_temporality={Counter: AggregationTemporality.DELTA}
        )
        meter_provider = MeterProvider(metric_readers=[reader])

        meter = meter_provider.get_meter("testmeter")

        counter_0 = meter.create_counter("counter_0")
        counter_1 = meter.create_counter("counter_1")

        counter_0.add(10, {"label": "value1"})
        sleep(self.delay)
        counter_0.add(10, {"label": "value2"})
        sleep(self.delay)
        counter_1.add(10, {"label": "value1"})
        sleep(self.delay)
        counter_1.add(10, {"label": "value2"})

        metrics = reader.get_metrics_data()

        data_points_0_0 = list(
            metrics.resource_metrics[0]
            .scope_metrics[0]
            .metrics[0]
            .data.data_points
        )
        data_points_0_1 = list(
            metrics.resource_metrics[0]
            .scope_metrics[0]
            .metrics[1]
            .data.data_points
        )
        self.assertEqual(len(data_points_0_0), 2)
        self.assertEqual(len(data_points_0_1), 2)

        self.assertLess(
            data_points_0_0[0].start_time_unix_nano,
            data_points_0_0[1].start_time_unix_nano,
        )
        self.assertLess(
            data_points_0_1[0].start_time_unix_nano,
            data_points_0_1[1].start_time_unix_nano,
        )
        self.assertNotEqual(
            data_points_0_0[1].start_time_unix_nano,
            data_points_0_1[0].start_time_unix_nano,
        )

        self.assertEqual(
            data_points_0_0[0].time_unix_nano,
            data_points_0_0[1].time_unix_nano,
        )
        self.assertEqual(
            data_points_0_1[0].time_unix_nano,
            data_points_0_1[1].time_unix_nano,
        )
        self.assertEqual(
            data_points_0_0[1].time_unix_nano,
            data_points_0_1[0].time_unix_nano,
        )

        counter_0.add(10, {"label": "value1"})
        sleep(self.delay)
        counter_0.add(10, {"label": "value2"})
        sleep(self.delay)
        counter_1.add(10, {"label": "value1"})
        sleep(self.delay)
        counter_1.add(10, {"label": "value2"})

        metrics = reader.get_metrics_data()

        data_points_1_0 = list(
            metrics.resource_metrics[0]
            .scope_metrics[0]
            .metrics[0]
            .data.data_points
        )
        data_points_1_1 = list(
            metrics.resource_metrics[0]
            .scope_metrics[0]
            .metrics[1]
            .data.data_points
        )
        self.assertEqual(len(data_points_1_0), 2)
        self.assertEqual(len(data_points_1_1), 2)

        self.assertEqual(
            data_points_1_0[0].start_time_unix_nano,
            data_points_1_0[1].start_time_unix_nano,
        )
        self.assertEqual(
            data_points_1_1[0].start_time_unix_nano,
            data_points_1_1[1].start_time_unix_nano,
        )
        self.assertEqual(
            data_points_1_0[1].start_time_unix_nano,
            data_points_1_1[0].start_time_unix_nano,
        )

        self.assertEqual(
            data_points_1_0[0].time_unix_nano,
            data_points_1_0[1].time_unix_nano,
        )
        self.assertEqual(
            data_points_1_1[0].time_unix_nano,
            data_points_1_1[1].time_unix_nano,
        )
        self.assertEqual(
            data_points_1_0[1].time_unix_nano,
            data_points_1_1[0].time_unix_nano,
        )

        self.assertNotEqual(
            data_points_0_0[0].start_time_unix_nano,
            data_points_1_0[0].start_time_unix_nano,
        )
        self.assertNotEqual(
            data_points_0_0[1].start_time_unix_nano,
            data_points_1_0[1].start_time_unix_nano,
        )
        self.assertNotEqual(
            data_points_0_1[0].start_time_unix_nano,
            data_points_1_1[0].start_time_unix_nano,
        )
        self.assertNotEqual(
            data_points_0_1[1].start_time_unix_nano,
            data_points_1_1[1].start_time_unix_nano,
        )

        self.assertEqual(
            data_points_0_0[0].time_unix_nano,
            data_points_1_0[0].start_time_unix_nano,
        )
        self.assertEqual(
            data_points_0_0[1].time_unix_nano,
            data_points_1_0[1].start_time_unix_nano,
        )
        self.assertEqual(
            data_points_0_1[0].time_unix_nano,
            data_points_1_1[0].start_time_unix_nano,
        )
        self.assertEqual(
            data_points_0_1[1].time_unix_nano,
            data_points_1_1[1].start_time_unix_nano,
        )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

"""Zipkin Exporter Transport Encoder

Base module and abstract class for concrete transport encoders to extend.
"""

import abc
import json
import logging
from collections.abc import Sequence
from enum import Enum
from typing import Any, TypeVar

from opentelemetry.exporter.zipkin.node_endpoint import NodeEndpoint
from opentelemetry.sdk.trace import Event
from opentelemetry.trace import (
    Span,
    SpanContext,
    StatusCode,
    format_span_id,
    format_trace_id,
)

EncodedLocalEndpointT = TypeVar("EncodedLocalEndpointT")

DEFAULT_MAX_TAG_VALUE_LENGTH = 128
NAME_KEY = "otel.library.name"
VERSION_KEY = "otel.library.version"
_SCOPE_NAME_KEY = "otel.scope.name"
_SCOPE_VERSION_KEY = "otel.scope.version"

logger = logging.getLogger(__name__)


class Protocol(Enum):
    """Enum of supported protocol formats.

    Values are human-readable strings so that they can be easily used by the
    OS environ var OTEL_EXPORTER_ZIPKIN_PROTOCOL (reserved for future usage).
    """

    V1 = "v1"
    V2 = "v2"


# pylint: disable=W0223
class Encoder(abc.ABC):
    """Base class for encoders that are used by the exporter.

    Args:
        max_tag_value_length: maximum length of an exported tag value. Values
          will be truncated to conform. Since values are serialized to a JSON
          list string, max_tag_value_length is honored at the element boundary.
    """

    def __init__(
        self, max_tag_value_length: int = DEFAULT_MAX_TAG_VALUE_LENGTH
    ):
        self.max_tag_value_length = max_tag_value_length

    @staticmethod
    @abc.abstractmethod
    def content_type() -> str:
        pass

    @abc.abstractmethod
    def serialize(
        self, spans: Sequence[Span], local_endpoint: NodeEndpoint
    ) -> str:
        pass

    @abc.abstractmethod
    def _encode_span(
        self, span: Span, encoded_local_endpoint: EncodedLocalEndpointT
    ) -> Any:
        """
        Per spec Zipkin fields that can be absent SHOULD be omitted from the
        payload when they are empty in the OpenTelemetry Span.

        https://github.com/open-telemetry/opentelemetry-specification/blob/master/specification/trace/sdk_exporters/zipkin.md#request-payload
        """

    @staticmethod
    @abc.abstractmethod
    def _encode_local_endpoint(
        local_endpoint: NodeEndpoint,
    ) -> EncodedLocalEndpointT:
        pass

    @staticmethod
    def _encode_debug(span_context) -> Any:
        return span_context.trace_flags.sampled

    @staticmethod
    @abc.abstractmethod
    def _encode_span_id(span_id: int) -> Any:
        pass

    @staticmethod
    @abc.abstractmethod
    def _encode_trace_id(trace_id: int) -> Any:
        pass

    @staticmethod
    def _get_parent_id(span_context) -> int | None:
        if isinstance(span_context, Span):
            parent_id = span_context.parent.span_id
        elif isinstance(span_context, SpanContext):
            parent_id = span_context.span_id
        else:
            parent_id = None
        return parent_id

    def _extract_tags_from_dict(
        self, tags_dict: dict | None
    ) -> dict[str, str]:
        tags = {}
        if not tags_dict:
            return tags
        for attribute_key, attribute_value in tags_dict.items():
            if isinstance(attribute_value, bool):
                value = str(attribute_value).lower()
            elif isinstance(attribute_value, (int, float, str)):
                value = str(attribute_value)
            elif isinstance(attribute_value, Sequence):
                value = self._extract_tag_value_string_from_sequence(
                    attribute_value
                )
                if not value:
                    logger.warning("Could not serialize tag %s", attribute_key)
                    continue
            else:
                logger.warning("Could not serialize tag %s", attribute_key)
                continue

            if (
                self.max_tag_value_length is not None
                and self.max_tag_value_length > 0
            ):
                value = value[: self.max_tag_value_length]
            tags[attribute_key] = value
        return tags

    def _extract_tag_value_string_from_sequence(self, sequence: Sequence):
        if self.max_tag_value_length and self.max_tag_value_length == 1:
            return None

        tag_value_elements = []
        running_string_length = (
            2  # accounts for array brackets in output string
        )
        defined_max_tag_value_length = (
            self.max_tag_value_length is not None
            and self.max_tag_value_length > 0
        )

        for element in sequence:
            if isinstance(element, bool):
                tag_value_element = str(element).lower()
            elif isinstance(element, (int, float, str)):
                tag_value_element = str(element)
            elif element is None:
                tag_value_element = None
            else:
                continue

            if defined_max_tag_value_length:
                if tag_value_element is None:
                    running_string_length += 4  # null with no quotes
                else:
                    # + 2 accounts for string quotation marks
                    running_string_length += len(tag_value_element) + 2

                if tag_value_elements:
                    # accounts for ',' item separator
                    running_string_length += 1

                if running_string_length > self.max_tag_value_length:
                    break

            tag_value_elements.append(tag_value_element)

        return json.dumps(tag_value_elements, separators=(",", ":"))

    def _extract_tags_from_span(self, span: Span) -> dict[str, str]:
        tags = self._extract_tags_from_dict(span.attributes)
        if span.resource:
            tags.update(self._extract_tags_from_dict(span.resource.attributes))
        if span.instrumentation_scope is not None:
            tags.update(
                {
                    NAME_KEY: span.instrumentation_scope.name,
                    VERSION_KEY: span.instrumentation_scope.version,
                    _SCOPE_NAME_KEY: span.instrumentation_scope.name,
                    _SCOPE_VERSION_KEY: span.instrumentation_scope.version,
                }
            )
        if span.status.status_code is not StatusCode.UNSET:
            tags.update({"otel.status_code": span.status.status_code.name})
            if span.status.status_code is StatusCode.ERROR:
                tags.update({"error": span.status.description or ""})

        if span.dropped_attributes:
            tags.update(
                {"otel.dropped_attributes_count": str(span.dropped_attributes)}
            )

        if span.dropped_events:
            tags.update(
                {"otel.dropped_events_count": str(span.dropped_events)}
            )

        if span.dropped_links:
            tags.update({"otel.dropped_links_count": str(span.dropped_links)})

        return tags

    def _extract_annotations_from_events(
        self, events: list[Event] | None
    ) -> list[dict] | None:
        if not events:
            return None

        annotations = []
        for event in events:
            attrs = {}
            for key, value in event.attributes.items():
                if (
                    isinstance(value, str)
                    and self.max_tag_value_length is not None
                    and self.max_tag_value_length > 0
                ):
                    value = value[: self.max_tag_value_length]
                attrs[key] = value

            annotations.append(
                {
                    "timestamp": self._nsec_to_usec_round(event.timestamp),
                    "value": json.dumps({event.name: attrs}, sort_keys=True),
                }
            )
        return annotations

    @staticmethod
    def _nsec_to_usec_round(nsec: int) -> int:
        """Round nanoseconds to microseconds

        Timestamp in zipkin spans is int of microseconds.
        See: https://zipkin.io/pages/instrumenting.html
        """
        return (nsec + 500) // 10**3


class JsonEncoder(Encoder):
    @staticmethod
    def content_type():
        return "application/json"

    def serialize(
        self, spans: Sequence[Span], local_endpoint: NodeEndpoint
    ) -> str:
        encoded_local_endpoint = self._encode_local_endpoint(local_endpoint)
        encoded_spans = []
        for span in spans:
            encoded_spans.append(
                self._encode_span(span, encoded_local_endpoint)
            )
        return json.dumps(encoded_spans)

    @staticmethod
    def _encode_local_endpoint(local_endpoint: NodeEndpoint) -> dict:
        encoded_local_endpoint = {"serviceName": local_endpoint.service_name}
        if local_endpoint.ipv4 is not None:
            encoded_local_endpoint["ipv4"] = str(local_endpoint.ipv4)
        if local_endpoint.ipv6 is not None:
            encoded_local_endpoint["ipv6"] = str(local_endpoint.ipv6)
        if local_endpoint.port is not None:
            encoded_local_endpoint["port"] = local_endpoint.port
        return encoded_local_endpoint

    @staticmethod
    def _encode_span_id(span_id: int) -> str:
        return format_span_id(span_id)

    @staticmethod
    def _encode_trace_id(trace_id: int) -> str:
        return format_trace_id(trace_id)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

OTEL_SDK_DISABLED = "OTEL_SDK_DISABLED"
"""
.. envvar:: OTEL_SDK_DISABLED

The :envvar:`OTEL_SDK_DISABLED` environment variable disables the SDK for all signals
Default: "false"
"""

OTEL_CONFIG_FILE = "OTEL_CONFIG_FILE"
"""
.. envvar:: OTEL_CONFIG_FILE

The :envvar:`OTEL_CONFIG_FILE` environment variable points the SDK at a
declarative configuration file (YAML or JSON). When set, the file is the
sole source for SDK construction. Spec-defined ``OTEL_*`` variables with
schema equivalents are ignored. Env vars may still be read indirectly by
components the file enables (e.g. resource detectors) and via
``${env:VAR}`` substitution inside the file.

Python-implementation extensions outside the spec (``OTEL_PYTHON_*``
variables such as ``OTEL_PYTHON_LOGGING_AUTO_INSTRUMENTATION_ENABLED`` or
:envvar:`OTEL_PYTHON_TRACER_CONFIGURATOR`) are also bypassed when
:envvar:`OTEL_CONFIG_FILE` is set, because the env-var initialisation path
is skipped entirely in favour of the declarative file. Honouring these
alongside a config file is tracked separately.

See the OpenTelemetry declarative configuration specification for details.
"""

OTEL_RESOURCE_ATTRIBUTES = "OTEL_RESOURCE_ATTRIBUTES"
"""
.. envvar:: OTEL_RESOURCE_ATTRIBUTES

The :envvar:`OTEL_RESOURCE_ATTRIBUTES` environment variable allows resource
attributes to be passed to the SDK at process invocation. The attributes from
:envvar:`OTEL_RESOURCE_ATTRIBUTES` are merged with those passed to
`Resource.create`, meaning :envvar:`OTEL_RESOURCE_ATTRIBUTES` takes *lower*
priority. Attributes should be in the format ``key1=value1,key2=value2``.
Additional details are available `in the specification
<https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/resource/sdk.md#specifying-resource-information-via-an-environment-variable>`__.

.. code-block:: console

    $ OTEL_RESOURCE_ATTRIBUTES="service.name=shoppingcard,will_be_overridden=foo" python - <<EOF
    import pprint
    from opentelemetry.sdk.resources import Resource
    pprint.pprint(Resource.create({"will_be_overridden": "bar"}).attributes)
    EOF
    {'service.name': 'shoppingcard',
    'telemetry.sdk.language': 'python',
    'telemetry.sdk.name': 'opentelemetry',
    'telemetry.sdk.version': '0.13.dev0',
    'will_be_overridden': 'bar'}
"""

OTEL_LOG_LEVEL = "OTEL_LOG_LEVEL"
"""
.. envvar:: OTEL_LOG_LEVEL

The :envvar:`OTEL_LOG_LEVEL` environment variable sets the log level used by the SDK logger
Default: "info"
"""

OTEL_TRACES_SAMPLER = "OTEL_TRACES_SAMPLER"
"""
.. envvar:: OTEL_TRACES_SAMPLER

The :envvar:`OTEL_TRACES_SAMPLER` environment variable sets the sampler to be used for traces.
Sampling is a mechanism to control the noise introduced by OpenTelemetry by reducing the number
of traces collected and sent to the backend
Default: "parentbased_always_on"
"""

OTEL_TRACES_SAMPLER_ARG = "OTEL_TRACES_SAMPLER_ARG"
"""
.. envvar:: OTEL_TRACES_SAMPLER_ARG

The :envvar:`OTEL_TRACES_SAMPLER_ARG` environment variable will only be used if OTEL_TRACES_SAMPLER is set.
Each Sampler type defines its own expected input, if any.
Invalid or unrecognized input is ignored,
i.e. the SDK behaves as if OTEL_TRACES_SAMPLER_ARG is not set.
"""

OTEL_BLRP_SCHEDULE_DELAY = "OTEL_BLRP_SCHEDULE_DELAY"
"""
.. envvar:: OTEL_BLRP_SCHEDULE_DELAY

The :envvar:`OTEL_BLRP_SCHEDULE_DELAY` represents the delay interval between two consecutive exports of the BatchLogRecordProcessor.
Default: 1000
"""

OTEL_BLRP_EXPORT_TIMEOUT = "OTEL_BLRP_EXPORT_TIMEOUT"
"""
.. envvar:: OTEL_BLRP_EXPORT_TIMEOUT

The :envvar:`OTEL_BLRP_EXPORT_TIMEOUT` represents the maximum allowed time to export data from the BatchLogRecordProcessor.
This environment variable currently does nothing, see https://github.com/open-telemetry/opentelemetry-python/issues/4555.
Default: 30000
"""

OTEL_BLRP_MAX_QUEUE_SIZE = "OTEL_BLRP_MAX_QUEUE_SIZE"
"""
.. envvar:: OTEL_BLRP_MAX_QUEUE_SIZE

The :envvar:`OTEL_BLRP_MAX_QUEUE_SIZE` represents the maximum queue size for the data export of the BatchLogRecordProcessor.
Default: 2048
"""

OTEL_BLRP_MAX_EXPORT_BATCH_SIZE = "OTEL_BLRP_MAX_EXPORT_BATCH_SIZE"
"""
.. envvar:: OTEL_BLRP_MAX_EXPORT_BATCH_SIZE

The :envvar:`OTEL_BLRP_MAX_EXPORT_BATCH_SIZE` represents the maximum batch size for the data export of the BatchLogRecordProcessor.
Default: 512
"""

OTEL_BSP_SCHEDULE_DELAY = "OTEL_BSP_SCHEDULE_DELAY"
"""
.. envvar:: OTEL_BSP_SCHEDULE_DELAY

The :envvar:`OTEL_BSP_SCHEDULE_DELAY` represents the delay interval between two consecutive exports of the BatchSpanProcessor.
Default: 5000
"""

OTEL_BSP_EXPORT_TIMEOUT = "OTEL_BSP_EXPORT_TIMEOUT"
"""
.. envvar:: OTEL_BSP_EXPORT_TIMEOUT

The :envvar:`OTEL_BSP_EXPORT_TIMEOUT` represents the maximum allowed time to export data from the BatchSpanProcessor.
Default: 30000
"""

OTEL_BSP_MAX_QUEUE_SIZE = "OTEL_BSP_MAX_QUEUE_SIZE"
"""
.. envvar:: OTEL_BSP_MAX_QUEUE_SIZE

The :envvar:`OTEL_BSP_MAX_QUEUE_SIZE` represents the maximum queue size for the data export of the BatchSpanProcessor.
Default: 2048
"""

OTEL_BSP_MAX_EXPORT_BATCH_SIZE = "OTEL_BSP_MAX_EXPORT_BATCH_SIZE"
"""
.. envvar:: OTEL_BSP_MAX_EXPORT_BATCH_SIZE

The :envvar:`OTEL_BSP_MAX_EXPORT_BATCH_SIZE` represents the maximum batch size for the data export of the BatchSpanProcessor.
Default: 512
"""

OTEL_ATTRIBUTE_COUNT_LIMIT = "OTEL_ATTRIBUTE_COUNT_LIMIT"
"""
.. envvar:: OTEL_ATTRIBUTE_COUNT_LIMIT

The :envvar:`OTEL_ATTRIBUTE_COUNT_LIMIT` represents the maximum allowed attribute count for spans, events and links.
This limit is overridden by model specific limits such as OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT.
Default: 128
"""

OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT = "OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT"
"""
.. envvar:: OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT

The :envvar:`OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT` represents the maximum allowed attribute length.
"""

OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT = "OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT"
"""
.. envvar:: OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT

The :envvar:`OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT` represents the maximum allowed event attribute count.
Default: 128
"""

OTEL_LINK_ATTRIBUTE_COUNT_LIMIT = "OTEL_LINK_ATTRIBUTE_COUNT_LIMIT"
"""
.. envvar:: OTEL_LINK_ATTRIBUTE_COUNT_LIMIT

The :envvar:`OTEL_LINK_ATTRIBUTE_COUNT_LIMIT` represents the maximum allowed link attribute count.
Default: 128
"""

OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT = "OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT"
"""
.. envvar:: OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT

The :envvar:`OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT` represents the maximum allowed span attribute count.
Default: 128
"""

OTEL_SPAN_ATTRIBUTE_VALUE_LENGTH_LIMIT = (
    "OTEL_SPAN_ATTRIBUTE_VALUE_LENGTH_LIMIT"
)
"""
.. envvar:: OTEL_SPAN_ATTRIBUTE_VALUE_LENGTH_LIMIT

The :envvar:`OTEL_SPAN_ATTRIBUTE_VALUE_LENGTH_LIMIT` represents the maximum allowed length
span attribute values can have. This takes precedence over :envvar:`OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT`.
"""

OTEL_LOGRECORD_ATTRIBUTE_COUNT_LIMIT = "OTEL_LOGRECORD_ATTRIBUTE_COUNT_LIMIT"
"""
.. envvar:: OTEL_LOGRECORD_ATTRIBUTE_COUNT_LIMIT

The :envvar:`OTEL_LOGRECORD_ATTRIBUTE_COUNT_LIMIT` represents the maximum allowed log record
attribute count. This takes precedence over :envvar:`OTEL_ATTRIBUTE_COUNT_LIMIT`.
"""

OTEL_LOGRECORD_ATTRIBUTE_VALUE_LENGTH_LIMIT = (
    "OTEL_LOGRECORD_ATTRIBUTE_VALUE_LENGTH_LIMIT"
)
"""
.. envvar:: OTEL_LOGRECORD_ATTRIBUTE_VALUE_LENGTH_LIMIT

The :envvar:`OTEL_LOGRECORD_ATTRIBUTE_VALUE_LENGTH_LIMIT` represents the maximum allowed length
log record attribute values can have. This takes precedence over
:envvar:`OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT`.
"""

OTEL_SPAN_EVENT_COUNT_LIMIT = "OTEL_SPAN_EVENT_COUNT_LIMIT"
"""
.. envvar:: OTEL_SPAN_EVENT_COUNT_LIMIT

The :envvar:`OTEL_SPAN_EVENT_COUNT_LIMIT` represents the maximum allowed span event count.
Default: 128
"""

OTEL_SPAN_LINK_COUNT_LIMIT = "OTEL_SPAN_LINK_COUNT_LIMIT"
"""
.. envvar:: OTEL_SPAN_LINK_COUNT_LIMIT

The :envvar:`OTEL_SPAN_LINK_COUNT_LIMIT` represents the maximum allowed span link count.
Default: 128
"""

OTEL_EXPORTER_ZIPKIN_ENDPOINT = "OTEL_EXPORTER_ZIPKIN_ENDPOINT"
"""
.. envvar:: OTEL_EXPORTER_ZIPKIN_ENDPOINT

Zipkin collector endpoint to which the exporter will send data. This may
include a path (e.g. ``http://example.com:9411/api/v2/spans``).
"""

OTEL_EXPORTER_ZIPKIN_TIMEOUT = "OTEL_EXPORTER_ZIPKIN_TIMEOUT"
"""
.. envvar:: OTEL_EXPORTER_ZIPKIN_TIMEOUT

Maximum time (in seconds) the Zipkin exporter will wait for each batch export.
Default: 10
"""

OTEL_EXPORTER_OTLP_PROTOCOL = "OTEL_EXPORTER_OTLP_PROTOCOL"
"""
.. envvar:: OTEL_EXPORTER_OTLP_PROTOCOL

The :envvar:`OTEL_EXPORTER_OTLP_PROTOCOL` represents the transport protocol for the
OTLP exporter.
"""

OTEL_EXPORTER_OTLP_TRACES_PROTOCOL = "OTEL_EXPORTER_OTLP_TRACES_PROTOCOL"
"""
.. envvar:: OTEL_EXPORTER_OTLP_TRACES_PROTOCOL

The :envvar:`OTEL_EXPORTER_OTLP_TRACES_PROTOCOL` represents the transport protocol for spans.
"""

OTEL_EXPORTER_OTLP_METRICS_PROTOCOL = "OTEL_EXPORTER_OTLP_METRICS_PROTOCOL"
"""
.. envvar:: OTEL_EXPORTER_OTLP_METRICS_PROTOCOL

The :envvar:`OTEL_EXPORTER_OTLP_METRICS_PROTOCOL` represents the transport protocol for metrics.
"""

OTEL_EXPORTER_OTLP_LOGS_PROTOCOL = "OTEL_EXPORTER_OTLP_LOGS_PROTOCOL"
"""
.. envvar:: OTEL_EXPORTER_OTLP_LOGS_PROTOCOL

The :envvar:`OTEL_EXPORTER_OTLP_LOGS_PROTOCOL` represents the transport protocol for logs.
"""

OTEL_EXPORTER_OTLP_CERTIFICATE = "OTEL_EXPORTER_OTLP_CERTIFICATE"
"""
.. envvar:: OTEL_EXPORTER_OTLP_CERTIFICATE

The :envvar:`OTEL_EXPORTER_OTLP_CERTIFICATE` stores the path to the certificate file for
TLS credentials of gRPC client. Should only be used for a secure connection.
"""

OTEL_EXPORTER_OTLP_HEADERS = "OTEL_EXPORTER_OTLP_HEADERS"
"""
.. envvar:: OTEL_EXPORTER_OTLP_HEADERS

The :envvar:`OTEL_EXPORTER_OTLP_HEADERS` contains the key-value pairs to be used as headers
associated with gRPC or HTTP requests.
"""


OTEL_EXPORTER_OTLP_COMPRESSION = "OTEL_EXPORTER_OTLP_COMPRESSION"
"""
.. envvar:: OTEL_EXPORTER_OTLP_COMPRESSION

Specifies a gRPC compression method to be used in the OTLP exporters.
Possible values are:

- ``gzip`` corresponding to `grpc.Compression.Gzip`.
- ``deflate`` corresponding to `grpc.Compression.Deflate`.

If no ``OTEL_EXPORTER_OTLP_*COMPRESSION`` environment variable is present or
``compression`` argument passed to the exporter, the default
`grpc.Compression.NoCompression` will be used. Additional details are
available `in the specification
<https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/protocol/exporter.md#opentelemetry-protocol-exporter>`__.
"""

OTEL_EXPORTER_OTLP_TIMEOUT = "OTEL_EXPORTER_OTLP_TIMEOUT"
"""
.. envvar:: OTEL_EXPORTER_OTLP_TIMEOUT

The :envvar:`OTEL_EXPORTER_OTLP_TIMEOUT` is the maximum time (in seconds) the OTLP exporter will wait for each batch export.
Default: 10
"""

OTEL_EXPORTER_OTLP_ENDPOINT = "OTEL_EXPORTER_OTLP_ENDPOINT"
"""
.. envvar:: OTEL_EXPORTER_OTLP_ENDPOINT

The :envvar:`OTEL_EXPORTER_OTLP_ENDPOINT` target to which the exporter is going to send spans or metrics.
The endpoint MUST be a valid URL host, and MAY contain a scheme (http or https), port and path.
A scheme of https indicates a secure connection and takes precedence over the insecure configuration setting.
Default: "http://localhost:4317"
"""

OTEL_EXPORTER_OTLP_INSECURE = "OTEL_EXPORTER_OTLP_INSECURE"
"""
.. envvar:: OTEL_EXPORTER_OTLP_INSECURE

The :envvar:`OTEL_EXPORTER_OTLP_INSECURE` represents whether to enable client transport security for gRPC requests.
A scheme of https takes precedence over this configuration setting.
Default: False
"""

OTEL_EXPORTER_OTLP_TRACES_INSECURE = "OTEL_EXPORTER_OTLP_TRACES_INSECURE"
"""
.. envvar:: OTEL_EXPORTER_OTLP_TRACES_INSECURE

The :envvar:`OTEL_EXPORTER_OTLP_TRACES_INSECURE` represents whether to enable client transport security
for gRPC requests for spans. A scheme of https takes precedence over the this configuration setting.
Default: False
"""


OTEL_EXPORTER_OTLP_TRACES_ENDPOINT = "OTEL_EXPORTER_OTLP_TRACES_ENDPOINT"
"""
.. envvar:: OTEL_EXPORTER_OTLP_TRACES_ENDPOINT

The :envvar:`OTEL_EXPORTER_OTLP_TRACES_ENDPOINT` target to which the span exporter is going to send spans.
The endpoint MUST be a valid URL host, and MAY contain a scheme (http or https), port and path.
A scheme of https indicates a secure connection and takes precedence over this configuration setting.
"""

OTEL_EXPORTER_OTLP_METRICS_ENDPOINT = "OTEL_EXPORTER_OTLP_METRICS_ENDPOINT"
"""
.. envvar:: OTEL_EXPORTER_OTLP_METRICS_ENDPOINT

The :envvar:`OTEL_EXPORTER_OTLP_METRICS_ENDPOINT` target to which the metrics exporter is going to send metrics.
The endpoint MUST be a valid URL host, and MAY contain a scheme (http or https), port and path.
A scheme of https indicates a secure connection and takes precedence over this configuration setting.
"""

OTEL_EXPORTER_OTLP_LOGS_ENDPOINT = "OTEL_EXPORTER_OTLP_LOGS_ENDPOINT"
"""
.. envvar:: OTEL_EXPORTER_OTLP_LOGS_ENDPOINT

The :envvar:`OTEL_EXPORTER_OTLP_LOGS_ENDPOINT` target to which the log exporter is going to send logs.
The endpoint MUST be a valid URL host, and MAY contain a scheme (http or https), port and path.
A scheme of https indicates a secure connection and takes precedence over this configuration setting.
"""

_OTEL_PYTHON_EXPORTER_OTLP_GRPC_LOGS_CREDENTIAL_PROVIDER = (
    "OTEL_PYTHON_EXPORTER_OTLP_GRPC_LOGS_CREDENTIAL_PROVIDER"
)
"""
.. envvar:: OTEL_PYTHON_EXPORTER_OTLP_GRPC_LOGS_CREDENTIAL_PROVIDER

The :envvar:`OTEL_PYTHON_EXPORTER_OTLP_GRPC_LOGS_CREDENTIAL_PROVIDER` provides `grpc.ChannelCredentials` to the grpc OTLP Log exporter,
Entry point providers should implement the following:

.. code-block:: python

    import grpc

    # Add a reference to this function under the `opentelemetry_otlp_credential_provider` entry point.
    def channel_credential_provider() -> grpc.ChannelCredentials:

Note: This environment variable is experimental and subject to change.
"""

_OTEL_PYTHON_EXPORTER_OTLP_HTTP_LOGS_CREDENTIAL_PROVIDER = (
    "OTEL_PYTHON_EXPORTER_OTLP_HTTP_LOGS_CREDENTIAL_PROVIDER"
)
"""
.. envvar:: OTEL_PYTHON_EXPORTER_OTLP_HTTP_LOGS_CREDENTIAL_PROVIDER

The :envvar:`OTEL_PYTHON_EXPORTER_OTLP_HTTP_LOGS_CREDENTIAL_PROVIDER` provides `requests.Session` for the HTTP OTLP Log exporter.
Entry point providers should implement the following:

.. code-block:: python

    import requests

    # Add a reference to this function under the `opentelemetry_otlp_credential_provider` entry point.
    def request_session_provder() -> requests.Session:

Note: This environment variable is experimental and subject to change.
"""
_OTEL_PYTHON_EXPORTER_OTLP_HTTP_CREDENTIAL_PROVIDER = (
    "OTEL_PYTHON_EXPORTER_OTLP_HTTP_CREDENTIAL_PROVIDER"
)
"""
.. envvar:: OTEL_PYTHON_EXPORTER_OTLP_HTTP_CREDENTIAL_PROVIDER

The :envvar:`OTEL_PYTHON_EXPORTER_OTLP_HTTP_CREDENTIAL_PROVIDER` provides `requests.Session` for all HTTP OTLP exporters.
Entry point providers should implement the following:

.. code-block:: python

    import requests

    # Add a reference to this function under the `opentelemetry_otlp_credential_provider` entry point.
    def request_session_provder() -> requests.Session:

Note: This environment variable is experimental and subject to change.
"""
_OTEL_PYTHON_EXPORTER_OTLP_GRPC_CREDENTIAL_PROVIDER = (
    "OTEL_PYTHON_EXPORTER_OTLP_GRPC_CREDENTIAL_PROVIDER"
)
"""
.. envvar:: OTEL_PYTHON_EXPORTER_OTLP_GRPC_CREDENTIAL_PROVIDER

The :envvar:`OTEL_PYTHON_EXPORTER_OTLP_GRPC_CREDENTIAL_PROVIDER` provides `grpc.ChannelCredentials` for all GRPC OTLP exporters.
Entry point providers should implement the following:

.. code-block:: python

    import grpc

    # Add a reference to this function under the `opentelemetry_otlp_credential_provider` entry point.
    def channel_credential_provider() -> grpc.ChannelCredentials:

Note: This environment variable is experimental and subject to change.
"""
_OTEL_PYTHON_EXPORTER_OTLP_HTTP_TRACES_CREDENTIAL_PROVIDER = (
    "OTEL_PYTHON_EXPORTER_OTLP_HTTP_TRACES_CREDENTIAL_PROVIDER"
)
"""
.. envvar:: OTEL_PYTHON_EXPORTER_OTLP_HTTP_TRACES_CREDENTIAL_PROVIDER

The :envvar:`OTEL_PYTHON_EXPORTER_OTLP_HTTP_TRACES_CREDENTIAL_PROVIDER` provides `requests.Session` to the HTTP OTLP Span exporter.
Entry point providers should implement the following:

.. code-block:: python

    import requests

    # Add a reference to this function under the `opentelemetry_otlp_credential_provider` entry point.
    def request_session_provder() -> requests.Session:

Note: This environment variable is experimental and subject to change.
"""
_OTEL_PYTHON_EXPORTER_OTLP_GRPC_TRACES_CREDENTIAL_PROVIDER = (
    "OTEL_PYTHON_EXPORTER_OTLP_GRPC_TRACES_CREDENTIAL_PROVIDER"
)
"""
.. envvar:: OTEL_PYTHON_EXPORTER_OTLP_GRPC_TRACES_CREDENTIAL_PROVIDER

The :envvar:`OTEL_PYTHON_EXPORTER_OTLP_GRPC_TRACES_CREDENTIAL_PROVIDER` provides `grpc.ChannelCredentials` to the GRPC OTLP Span exporter.
Entry point providers should implement the following:

.. code-block:: python

    import grpc

    # Add a reference to this function under the `opentelemetry_otlp_credential_provider` entry point.
    def channel_credential_provider() -> grpc.ChannelCredentials:

Note: This environment variable is experimental and subject to change.
"""
_OTEL_PYTHON_EXPORTER_OTLP_HTTP_METRICS_CREDENTIAL_PROVIDER = (
    "OTEL_PYTHON_EXPORTER_OTLP_HTTP_METRICS_CREDENTIAL_PROVIDER"
)
"""
.. envvar:: OTEL_PYTHON_EXPORTER_OTLP_HTTP_METRICS_CREDENTIAL_PROVIDER

The :envvar:`OTEL_PYTHON_EXPORTER_OTLP_HTTP_METRICS_CREDENTIAL_PROVIDER` provides `requests.Session` to the HTTP OTLP Metric exporter.
Entry point providers should implement the following:

.. code-block:: python

    import requests

    # Add a reference to this function under the `opentelemetry_otlp_credential_provider` entry point.
    def request_session_provder() -> requests.Session:

Note: This environment variable is experimental and subject to change.
"""
_OTEL_PYTHON_EXPORTER_OTLP_GRPC_METRICS_CREDENTIAL_PROVIDER = (
    "OTEL_PYTHON_EXPORTER_OTLP_GRPC_METRICS_CREDENTIAL_PROVIDER"
)
"""
.. envvar:: OTEL_PYTHON_EXPORTER_OTLP_GRPC_METRICS_CREDENTIAL_PROVIDER

The :envvar:`OTEL_PYTHON_EXPORTER_OTLP_GRPC_METRICS_CREDENTIAL_PROVIDER` provides `grpc.ChannelCredentials` to the GRPC OTLP Metric exporter.
Entry point providers should implement the following:

.. code-block:: python

    import grpc

    # Add a reference to this function under the `opentelemetry_otlp_credential_provider` entry point.
    def channel_credential_provider() -> grpc.ChannelCredentials:

Note: This environment variable is experimental and subject to change.
"""

_OTEL_PYTHON_EXPORTER_OTLP_GRPC_RETRYABLE_ERROR_CODES = (
    "OTEL_PYTHON_EXPORTER_OTLP_GRPC_RETRYABLE_ERROR_CODES"
)
"""
.. envvar:: OTEL_PYTHON_EXPORTER_OTLP_GRPC_RETRYABLE_ERROR_CODES

The :envvar:`OTEL_PYTHON_EXPORTER_OTLP_GRPC_RETRYABLE_ERROR_CODES` stores a comma-separated list of human-readable
gRPC error codes that are considered retryable for the OTLP gRPC exporters (e.g. `UNAVAILABLE, DEADLINE_EXCEEDED`).
Supported error codes are defined in `grpc.StatusCode` and are parsed in a case-insensitive manner.

Note: This environment variable is experimental and subject to change.
"""

OTEL_EXPORTER_OTLP_TRACES_CERTIFICATE = "OTEL_EXPORTER_OTLP_TRACES_CERTIFICATE"
"""
.. envvar:: OTEL_EXPORTER_OTLP_TRACES_CERTIFICATE

The :envvar:`OTEL_EXPORTER_OTLP_TRACES_CERTIFICATE` stores the path to the certificate file for
TLS credentials of gRPC client for traces. Should only be used for a secure connection for tracing.
"""

OTEL_EXPORTER_OTLP_METRICS_CERTIFICATE = (
    "OTEL_EXPORTER_OTLP_METRICS_CERTIFICATE"
)
"""
.. envvar:: OTEL_EXPORTER_OTLP_METRICS_CERTIFICATE

The :envvar:`OTEL_EXPORTER_OTLP_METRICS_CERTIFICATE` stores the path to the certificate file for
TLS credentials of gRPC client for metrics. Should only be used for a secure connection for exporting metrics.
"""

OTEL_EXPORTER_OTLP_CLIENT_KEY = "OTEL_EXPORTER_OTLP_CLIENT_KEY"
"""
.. envvar:: OTEL_EXPORTER_OTLP_CLIENT_KEY

The :envvar:`OTEL_EXPORTER_OTLP_CLIENT_KEY` stores the path to the client private key to use
in mTLS communication in PEM format.
"""

OTEL_EXPORTER_OTLP_TRACES_CLIENT_KEY = "OTEL_EXPORTER_OTLP_TRACES_CLIENT_KEY"
"""
.. envvar:: OTEL_EXPORTER_OTLP_TRACES_CLIENT_KEY

The :envvar:`OTEL_EXPORTER_OTLP_TRACES_CLIENT_KEY` stores the path to the client private key to use
in mTLS communication in PEM format for traces.
"""

OTEL_EXPORTER_OTLP_METRICS_CLIENT_KEY = "OTEL_EXPORTER_OTLP_METRICS_CLIENT_KEY"
"""
.. envvar:: OTEL_EXPORTER_OTLP_METRICS_CLIENT_KEY

The :envvar:`OTEL_EXPORTER_OTLP_METRICS_CLIENT_KEY` stores the path to the client private key to use
in mTLS communication in PEM format for metrics.
"""

OTEL_EXPORTER_OTLP_LOGS_CLIENT_KEY = "OTEL_EXPORTER_OTLP_LOGS_CLIENT_KEY"
"""
.. envvar:: OTEL_EXPORTER_OTLP_LOGS_CLIENT_KEY

The :envvar:`OTEL_EXPORTER_OTLP_LOGS_CLIENT_KEY` stores the path to the client private key to use
in mTLS communication in PEM format for logs.
"""

OTEL_EXPORTER_OTLP_CLIENT_CERTIFICATE = "OTEL_EXPORTER_OTLP_CLIENT_CERTIFICATE"
"""
.. envvar:: OTEL_EXPORTER_OTLP_CLIENT_CERTIFICATE

The :envvar:`OTEL_EXPORTER_OTLP_CLIENT_CERTIFICATE` stores the path to the client certificate/chain trust for
clients private key to use in mTLS communication in PEM format.
"""

OTEL_EXPORTER_OTLP_TRACES_CLIENT_CERTIFICATE = (
    "OTEL_EXPORTER_OTLP_TRACES_CLIENT_CERTIFICATE"
)
"""
.. envvar:: OTEL_EXPORTER_OTLP_TRACES_CLIENT_CERTIFICATE

The :envvar:`OTEL_EXPORTER_OTLP_TRACES_CLIENT_CERTIFICATE` stores the path to the client certificate/chain trust for
clients private key to use in mTLS communication in PEM format for traces.
"""

OTEL_EXPORTER_OTLP_METRICS_CLIENT_CERTIFICATE = (
    "OTEL_EXPORTER_OTLP_METRICS_CLIENT_CERTIFICATE"
)
"""
.. envvar:: OTEL_EXPORTER_OTLP_METRICS_CLIENT_CERTIFICATE

The :envvar:`OTEL_EXPORTER_OTLP_METRICS_CLIENT_CERTIFICATE` stores the path to the client certificate/chain trust for
clients private key to use in mTLS communication in PEM format for metrics.
"""

OTEL_EXPORTER_OTLP_LOGS_CLIENT_CERTIFICATE = (
    "OTEL_EXPORTER_OTLP_LOGS_CLIENT_CERTIFICATE"
)
"""
.. envvar:: OTEL_EXPORTER_OTLP_LOGS_CLIENT_CERTIFICATE

The :envvar:`OTEL_EXPORTER_OTLP_LOGS_CLIENT_CERTIFICATE` stores the path to the client certificate/chain trust for
clients private key to use in mTLS communication in PEM format for logs.
"""

OTEL_EXPORTER_OTLP_TRACES_HEADERS = "OTEL_EXPORTER_OTLP_TRACES_HEADERS"
"""
.. envvar:: OTEL_EXPORTER_OTLP_TRACES_HEADERS

The :envvar:`OTEL_EXPORTER_OTLP_TRACES_HEADERS` contains the key-value pairs to be used as headers for spans
associated with gRPC or HTTP requests.
"""

OTEL_EXPORTER_OTLP_METRICS_HEADERS = "OTEL_EXPORTER_OTLP_METRICS_HEADERS"
"""
.. envvar:: OTEL_EXPORTER_OTLP_METRICS_HEADERS

The :envvar:`OTEL_EXPORTER_OTLP_METRICS_HEADERS` contains the key-value pairs to be used as headers for metrics
associated with gRPC or HTTP requests.
"""

OTEL_EXPORTER_OTLP_LOGS_HEADERS = "OTEL_EXPORTER_OTLP_LOGS_HEADERS"
"""
.. envvar:: OTEL_EXPORTER_OTLP_LOGS_HEADERS

The :envvar:`OTEL_EXPORTER_OTLP_LOGS_HEADERS` contains the key-value pairs to be used as headers for logs
associated with gRPC or HTTP requests.
"""

OTEL_EXPORTER_OTLP_TRACES_COMPRESSION = "OTEL_EXPORTER_OTLP_TRACES_COMPRESSION"
"""
.. envvar:: OTEL_EXPORTER_OTLP_TRACES_COMPRESSION

Same as :envvar:`OTEL_EXPORTER_OTLP_COMPRESSION` but only for the span
exporter. If both are present, this takes higher precedence.
"""

OTEL_EXPORTER_OTLP_METRICS_COMPRESSION = (
    "OTEL_EXPORTER_OTLP_METRICS_COMPRESSION"
)
"""
.. envvar:: OTEL_EXPORTER_OTLP_METRICS_COMPRESSION

Same as :envvar:`OTEL_EXPORTER_OTLP_COMPRESSION` but only for the metric
exporter. If both are present, this takes higher precedence.
"""

OTEL_EXPORTER_OTLP_LOGS_COMPRESSION = "OTEL_EXPORTER_OTLP_LOGS_COMPRESSION"
"""
.. envvar:: OTEL_EXPORTER_OTLP_LOGS_COMPRESSION

Same as :envvar:`OTEL_EXPORTER_OTLP_COMPRESSION` but only for the log
exporter. If both are present, this takes higher precedence.
"""

OTEL_EXPORTER_OTLP_TRACES_TIMEOUT = "OTEL_EXPORTER_OTLP_TRACES_TIMEOUT"
"""
.. envvar:: OTEL_EXPORTER_OTLP_TRACES_TIMEOUT

The :envvar:`OTEL_EXPORTER_OTLP_TRACES_TIMEOUT` is the maximum time (in seconds) the OTLP exporter will
wait for each batch export for spans.
Default: 10
"""

OTEL_EXPORTER_OTLP_METRICS_TIMEOUT = "OTEL_EXPORTER_OTLP_METRICS_TIMEOUT"
"""
.. envvar:: OTEL_EXPORTER_OTLP_METRICS_TIMEOUT

The :envvar:`OTEL_EXPORTER_OTLP_METRICS_TIMEOUT` is the maximum time (in seconds) the OTLP exporter will
wait for each batch export for metrics.
Default: 10
"""

OTEL_EXPORTER_OTLP_METRICS_INSECURE = "OTEL_EXPORTER_OTLP_METRICS_INSECURE"
"""
.. envvar:: OTEL_EXPORTER_OTLP_METRICS_INSECURE

The :envvar:`OTEL_EXPORTER_OTLP_METRICS_INSECURE` represents whether to enable client transport security
for gRPC requests for metrics. A scheme of https takes precedence over the this configuration setting.
Default: False
"""

OTEL_EXPORTER_OTLP_LOGS_INSECURE = "OTEL_EXPORTER_OTLP_LOGS_INSECURE"
"""
.. envvar:: OTEL_EXPORTER_OTLP_LOGS_INSECURE

The :envvar:`OTEL_EXPORTER_OTLP_LOGS_INSECURE` represents whether to enable client transport security
for gRPC requests for logs. A scheme of https takes precedence over the this configuration setting.
Default: False
"""

OTEL_EXPORTER_OTLP_LOGS_CERTIFICATE = "OTEL_EXPORTER_OTLP_LOGS_CERTIFICATE"
"""
.. envvar:: OTEL_EXPORTER_OTLP_LOGS_CERTIFICATE

The :envvar:`OTEL_EXPORTER_OTLP_LOGS_CERTIFICATE` stores the path to the certificate file for
TLS credentials of gRPC client for logs. Should only be used for a secure connection for logs.
"""

OTEL_EXPORTER_OTLP_LOGS_TIMEOUT = "OTEL_EXPORTER_OTLP_LOGS_TIMEOUT"
"""
.. envvar:: OTEL_EXPORTER_OTLP_LOGS_TIMEOUT

The :envvar:`OTEL_EXPORTER_OTLP_LOGS_TIMEOUT` is the maximum time (in seconds) the OTLP exporter will
wait for each batch export for logs.
Default: 10
"""

OTEL_SERVICE_NAME = "OTEL_SERVICE_NAME"
"""
.. envvar:: OTEL_SERVICE_NAME

Convenience environment variable for setting the service name resource attribute.
The following two environment variables have the same effect

.. code-block:: console

    OTEL_SERVICE_NAME=my-python-service

    OTEL_RESOURCE_ATTRIBUTES=service.name=my-python-service


If both are set, :envvar:`OTEL_SERVICE_NAME` takes precedence.
"""


_OTEL_PYTHON_LOGGING_AUTO_INSTRUMENTATION_ENABLED = (
    "OTEL_PYTHON_LOGGING_AUTO_INSTRUMENTATION_ENABLED"
)
"""
.. envvar:: OTEL_PYTHON_LOGGING_AUTO_INSTRUMENTATION_ENABLED

The :envvar:`OTEL_PYTHON_LOGGING_AUTO_INSTRUMENTATION_ENABLED` environment variable allows users to
enable/disable the auto instrumentation for the python logging module.
Default: False

Note: Logs SDK and its related settings are experimental.

.. warning::

    This option is deprecated, instead you should install `opentelemetry-instrumentation-logging`.
"""


OTEL_EXPORTER_OTLP_METRICS_TEMPORALITY_PREFERENCE = (
    "OTEL_EXPORTER_OTLP_METRICS_TEMPORALITY_PREFERENCE"
)
"""
.. envvar:: OTEL_EXPORTER_OTLP_METRICS_TEMPORALITY_PREFERENCE

The :envvar:`OTEL_EXPORTER_OTLP_METRICS_TEMPORALITY_PREFERENCE` environment
variable allows users to set the default aggregation temporality policy to use
on the basis of instrument kind. The valid (case-insensitive) values are:

``CUMULATIVE``: Use ``CUMULATIVE`` aggregation temporality for all instrument kinds.
``DELTA``: Use ``DELTA`` aggregation temporality for ``Counter``, ``Asynchronous Counter`` and ``Histogram``.
Use ``CUMULATIVE`` aggregation temporality for ``UpDownCounter`` and ``Asynchronous UpDownCounter``.
``LOWMEMORY``: Use ``DELTA`` aggregation temporality for ``Counter`` and ``Histogram``.
Use ``CUMULATIVE`` aggregation temporality for ``UpDownCounter``, ``AsynchronousCounter`` and ``Asynchronous UpDownCounter``.
"""

OTEL_METRIC_EXPORT_INTERVAL = "OTEL_METRIC_EXPORT_INTERVAL"
"""
.. envvar:: OTEL_METRIC_EXPORT_INTERVAL

The :envvar:`OTEL_METRIC_EXPORT_INTERVAL` is the time interval (in milliseconds) between the start of two export attempts.
"""

OTEL_METRIC_EXPORT_TIMEOUT = "OTEL_METRIC_EXPORT_TIMEOUT"
"""
.. envvar:: OTEL_METRIC_EXPORT_TIMEOUT

The :envvar:`OTEL_METRIC_EXPORT_TIMEOUT` is the maximum allowed time (in milliseconds) to export data.
"""

OTEL_METRICS_EXEMPLAR_FILTER = "OTEL_METRICS_EXEMPLAR_FILTER"
"""
.. envvar:: OTEL_METRICS_EXEMPLAR_FILTER

The :envvar:`OTEL_METRICS_EXEMPLAR_FILTER` is the filter for which measurements can become Exemplars.
"""

OTEL_EXPORTER_OTLP_METRICS_DEFAULT_HISTOGRAM_AGGREGATION = (
    "OTEL_EXPORTER_OTLP_METRICS_DEFAULT_HISTOGRAM_AGGREGATION"
)
"""
.. envvar:: OTEL_EXPORTER_OTLP_METRICS_DEFAULT_HISTOGRAM_AGGREGATION

The :envvar:`OTEL_EXPORTER_OTLP_METRICS_DEFAULT_HISTOGRAM_AGGREGATION` is the default aggregation to use for histogram instruments.
"""

OTEL_EXPERIMENTAL_RESOURCE_DETECTORS = "OTEL_EXPERIMENTAL_RESOURCE_DETECTORS"
"""
.. envvar:: OTEL_EXPERIMENTAL_RESOURCE_DETECTORS

The :envvar:`OTEL_EXPERIMENTAL_RESOURCE_DETECTORS` is a comma-separated string
of names of resource detectors. These names must be the same as the names of
entry points for the ```opentelemetry_resource_detector``` entry point. This is an
experimental feature and the name of this variable and its behavior can change
in a non-backwards compatible way.

Detectors are run in the order they are listed and their attributes are merged
in that order, with later detectors taking precedence over earlier ones on
conflicting keys. The ``otel`` detector (which reads
:envvar:`OTEL_RESOURCE_ATTRIBUTES` and :envvar:`OTEL_SERVICE_NAME`) is always
appended last unless explicitly placed elsewhere in the list, ensuring
environment variable attributes take highest priority among detectors.
"""

OTEL_EXPORTER_PROMETHEUS_HOST = "OTEL_EXPORTER_PROMETHEUS_HOST"
"""
.. envvar:: OTEL_EXPORTER_PROMETHEUS_HOST

The :envvar:`OTEL_EXPORTER_PROMETHEUS_HOST` environment variable configures the host used by
the Prometheus exporter.
Default: "localhost"

This is an experimental environment variable and the name of this variable and its behavior can
change in a non-backwards compatible way.
"""

OTEL_EXPORTER_PROMETHEUS_PORT = "OTEL_EXPORTER_PROMETHEUS_PORT"
"""
.. envvar:: OTEL_EXPORTER_PROMETHEUS_PORT

The :envvar:`OTEL_EXPORTER_PROMETHEUS_PORT` environment variable configures the port used by
the Prometheus exporter.
Default: 9464

This is an experimental environment variable and the name of this variable and its behavior can
change in a non-backwards compatible way.
"""

OTEL_PYTHON_TRACER_CONFIGURATOR = "OTEL_PYTHON_TRACER_CONFIGURATOR"
"""
.. envvar:: OTEL_PYTHON_TRACER_CONFIGURATOR

The :envvar:`OTEL_PYTHON_TRACER_CONFIGURATOR` environment variable allows users to set a
custom Tracer Configurator function.
Default: opentelemetry.sdk.trace._default_tracer_configurator

This is an experimental environment variable and the name of this variable and its behavior can
change in a non-backwards compatible way.
"""

OTEL_PYTHON_METER_CONFIGURATOR = "OTEL_PYTHON_METER_CONFIGURATOR"
"""
.. envvar:: OTEL_PYTHON_METER_CONFIGURATOR

The :envvar:`OTEL_PYTHON_METER_CONFIGURATOR` environment variable allows users to set a
custom Meter Configurator function.
Default: opentelemetry.sdk.metrics._internal._default_meter_configurator

This is an experimental environment variable and the name of this variable and its behavior can
change in a non-backwards compatible way.
"""

OTEL_PYTHON_LOGGER_CONFIGURATOR = "OTEL_PYTHON_LOGGER_CONFIGURATOR"
"""
.. envvar:: OTEL_PYTHON_LOGGER_CONFIGURATOR

The :envvar:`OTEL_PYTHON_LOGGER_CONFIGURATOR` environment variable allows users to set a
custom Logger Configurator function.
Default: opentelemetry.sdk._logs._internal._default_logger_configurator

This is an experimental environment variable and the name of this variable and its behavior can
change in a non-backwards compatible way.
"""

OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED = (
    "OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED"
)
"""
.. envvar:: OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED

The :envvar:`OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED` environment variable enables
metrics emitted by the SDK about its own internal state.
Default: "false"
"""



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

"""Zipkin Export Encoders for JSON formats"""

from opentelemetry.exporter.zipkin.encoder import JsonEncoder
from opentelemetry.trace import Span, SpanKind


class JsonV2Encoder(JsonEncoder):
    """Zipkin Export Encoder for JSON v2 API

    API spec: https://github.com/openzipkin/zipkin-api/blob/master/zipkin2-api.yaml
    """

    SPAN_KIND_MAP = {
        SpanKind.INTERNAL: None,
        SpanKind.SERVER: "SERVER",
        SpanKind.CLIENT: "CLIENT",
        SpanKind.PRODUCER: "PRODUCER",
        SpanKind.CONSUMER: "CONSUMER",
    }

    def _encode_span(self, span: Span, encoded_local_endpoint: dict) -> dict:
        context = span.get_span_context()
        encoded_span = {
            "traceId": self._encode_trace_id(context.trace_id),
            "id": self._encode_span_id(context.span_id),
            "name": span.name,
            "timestamp": self._nsec_to_usec_round(span.start_time),
            "duration": self._nsec_to_usec_round(
                span.end_time - span.start_time
            ),
            "localEndpoint": encoded_local_endpoint,
            "kind": self.SPAN_KIND_MAP[span.kind],
        }

        tags = self._extract_tags_from_span(span)
        if tags:
            encoded_span["tags"] = tags

        annotations = self._extract_annotations_from_events(span.events)
        if annotations:
            encoded_span["annotations"] = annotations

        debug = self._encode_debug(context)
        if debug:
            encoded_span["debug"] = debug

        parent_id = self._get_parent_id(span.parent)
        if parent_id is not None:
            encoded_span["parentId"] = self._encode_span_id(parent_id)

        return encoded_span



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0
# pylint: disable=import-error

import unittest
from unittest.mock import MagicMock, patch

from opentelemetry.exporter.http.transport import _load_http_transport_factory
from opentelemetry.exporter.http.transport._requests import (
    RequestsHTTPTransport,
)
from opentelemetry.exporter.http.transport._urllib3 import (
    Urllib3HTTPTransport,
)

_ENTRY_POINTS_TARGET = "opentelemetry.util._importlib_metadata.entry_points"


# pylint: disable=no-self-use
class TestLoadHTTPTransportFactory(unittest.TestCase):
    def test_returns_requests_transport(self):
        self.assertIs(
            _load_http_transport_factory("requests"), RequestsHTTPTransport
        )

    def test_returns_urllib3_transport(self):
        self.assertIs(
            _load_http_transport_factory("urllib3"), Urllib3HTTPTransport
        )

    def test_known_transport_does_not_call_entry_points(self):
        with patch(_ENTRY_POINTS_TARGET) as mock_ep:
            _load_http_transport_factory("requests")
            _load_http_transport_factory("urllib3")
        self.assertFalse(mock_ep.called)

    def test_unknown_transport_calls_entry_points(self):
        def _custom_factory(*, verify, cert, **kwargs):
            pass

        mock_ep = MagicMock()
        mock_ep.load.return_value = _custom_factory
        with patch(_ENTRY_POINTS_TARGET, return_value=[mock_ep]) as mock_fn:
            result = _load_http_transport_factory("custom")
        self.assertEqual(
            mock_fn.call_args.kwargs,
            {"group": "opentelemetry_http_transport", "name": "custom"},
        )
        self.assertIs(result, _custom_factory)

    def test_entry_point_non_callable_raises_type_error(self):
        mock_ep = MagicMock()
        mock_ep.load.return_value = "not_callable"
        with patch(_ENTRY_POINTS_TARGET, return_value=[mock_ep]):
            self.assertRaises(TypeError, _load_http_transport_factory, "bad")

    def test_unknown_transport_raises_value_error(self):
        with patch(_ENTRY_POINTS_TARGET, return_value=[]):
            self.assertRaises(
                ValueError, _load_http_transport_factory, "nonexistent"
            )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import os
import unittest
from unittest.mock import patch

import yaml

from opentelemetry.configuration.file import (
    EnvSubstitutionError,
    substitute_env_vars,
)


class TestEnvSubstitution(unittest.TestCase):
    """Test environment variable substitution."""

    def test_simple_substitution(self):
        """Test basic ${VAR} substitution."""
        with patch.dict(os.environ, {"SERVICE_NAME": "my-service"}):
            result = substitute_env_vars("name: ${SERVICE_NAME}")
            self.assertEqual(result, "name: my-service")

    def test_multiple_substitutions(self):
        """Test multiple ${VAR} in same string."""
        with patch.dict(os.environ, {"HOST": "localhost", "PORT": "8080"}):
            result = substitute_env_vars("url: http://${HOST}:${PORT}")
            self.assertEqual(result, "url: http://localhost:8080")

    def test_substitution_with_default(self):
        """Test ${VAR:-default} with missing variable."""
        with patch.dict(os.environ, {}, clear=True):
            result = substitute_env_vars("name: ${SERVICE_NAME:-default}")
            self.assertEqual(result, "name: default")

    def test_substitution_with_default_override(self):
        """Test ${VAR:-default} with present variable."""
        with patch.dict(os.environ, {"SERVICE_NAME": "actual"}):
            result = substitute_env_vars("name: ${SERVICE_NAME:-default}")
            self.assertEqual(result, "name: actual")

    def test_missing_variable_raises_error(self):
        """Test ${VAR} raises error when variable missing."""
        with patch.dict(os.environ, {}, clear=True):
            with self.assertRaises(EnvSubstitutionError) as ctx:
                substitute_env_vars("name: ${MISSING_VAR}")
            self.assertIn("MISSING_VAR", str(ctx.exception))

    def test_dollar_sign_escape(self):
        """Test $$ escapes to literal $."""
        result = substitute_env_vars("price: $$100")
        self.assertEqual(result, "price: $100")

    def test_dollar_sign_escape_with_substitution(self):
        """Test $$ with ${VAR} in same string."""
        with patch.dict(os.environ, {"AMOUNT": "50"}):
            result = substitute_env_vars("price: $$${AMOUNT}")
            self.assertEqual(result, "price: $50")

    def test_empty_default(self):
        """Test ${VAR:-} with empty default value."""
        with patch.dict(os.environ, {}, clear=True):
            result = substitute_env_vars("name: ${VAR:-}")
            self.assertEqual(result, "name: ")

    def test_default_with_special_chars(self):
        """Test default value containing special characters."""
        with patch.dict(os.environ, {}, clear=True):
            result = substitute_env_vars("url: ${URL:-http://localhost:8080}")
            self.assertEqual(result, "url: http://localhost:8080")

    def test_no_substitution_needed(self):
        """Test string without any variables."""
        result = substitute_env_vars("plain text without variables")
        self.assertEqual(result, "plain text without variables")

    def test_variable_name_with_underscores(self):
        """Test variable names with underscores."""
        with patch.dict(os.environ, {"MY_VAR_NAME": "value"}):
            result = substitute_env_vars("key: ${MY_VAR_NAME}")
            self.assertEqual(result, "key: value")

    def test_variable_name_with_numbers(self):
        """Test variable names with numbers."""
        with patch.dict(os.environ, {"VAR_123": "value"}):
            result = substitute_env_vars("key: ${VAR_123}")
            self.assertEqual(result, "key: value")

    def test_multiline_substitution(self):
        """Test substitution across multiple lines."""
        with patch.dict(os.environ, {"VAR1": "value1", "VAR2": "value2"}):
            text = """line1: ${VAR1}
line2: ${VAR2}"""
            result = substitute_env_vars(text)
            expected = """line1: value1
line2: value2"""
            self.assertEqual(result, expected)

    def test_empty_string(self):
        """Test empty string returns empty string."""
        result = substitute_env_vars("")
        self.assertEqual(result, "")

    def test_only_dollar_signs(self):
        """Test string with only escaped dollar signs."""
        result = substitute_env_vars("$$$$")
        self.assertEqual(result, "$$")

    def test_newline_in_value_prevents_yaml_injection(self):
        """Values containing newlines must not inject YAML structure.

        Per spec: "It MUST NOT be possible to inject YAML structures by
        environment variables." A value like "legit\\nmalicious_key: val"
        must be emitted as a quoted scalar, not raw YAML.
        """
        with patch.dict(
            os.environ,
            {"SERVICE_NAME": "legit-service\nmalicious_key: injected_value"},
        ):
            result = substitute_env_vars(
                "file_format: '1.0'\nservice_name: ${SERVICE_NAME}"
            )
        parsed = yaml.safe_load(result)
        self.assertNotIn("malicious_key", parsed)
        self.assertIn("legit-service", parsed["service_name"])

    def test_newline_in_value_preserved_as_literal(self):
        """Newline within a value is preserved as a literal newline character."""
        with patch.dict(os.environ, {"MULTI": "line1\nline2"}):
            result = substitute_env_vars("key: ${MULTI}")
        parsed = yaml.safe_load(result)
        self.assertEqual(parsed["key"], "line1\nline2")

    def test_carriage_return_in_value_is_escaped(self):
        """Carriage return in value is escaped, not injected."""
        with patch.dict(os.environ, {"VAL": "text\r\nmore"}):
            result = substitute_env_vars("key: ${VAL}")
        parsed = yaml.safe_load(result)
        self.assertIsInstance(parsed["key"], str)

    def test_type_coercion_preserved_for_simple_values(self):
        """Simple values without newlines still undergo YAML type coercion per spec."""
        with patch.dict(os.environ, {"BOOL_VAL": "true", "INT_VAL": "42"}):
            bool_result = yaml.safe_load(
                substitute_env_vars("key: ${BOOL_VAL}")
            )
            int_result = yaml.safe_load(substitute_env_vars("key: ${INT_VAL}"))
        self.assertIs(bool_result["key"], True)
        self.assertEqual(int_result["key"], 42)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0


class IterEntryPoint:
    def __init__(self, name, class_type):
        self.name = name
        self.class_type = class_type

    def load(self):
        return self.class_type



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import json
import os

from opentelemetry.sdk import trace
from opentelemetry.sdk.resources import PROCESS_PID


def main() -> None:
    tracer_provider = trace.TracerProvider(shutdown_on_exit=False)
    tracer = tracer_provider.get_tracer("cached")
    parent_pid = os.getpid()
    parent_resource_pid = tracer_provider.resource.attributes[PROCESS_PID]
    parent_tracer_pid = tracer.resource.attributes[PROCESS_PID]

    pid = os.fork()
    if not pid:
        child_pid = os.getpid()
        new_tracer = tracer_provider.get_tracer("new")
        span = tracer.start_span("child")
        print(
            json.dumps(
                {
                    "child_pid": child_pid,
                    "provider_pid": tracer_provider.resource.attributes[
                        PROCESS_PID
                    ],
                    "cached_tracer_pid": tracer.resource.attributes[
                        PROCESS_PID
                    ],
                    "new_tracer_pid": new_tracer.resource.attributes[
                        PROCESS_PID
                    ],
                    "span_pid": span.resource.attributes[PROCESS_PID],
                }
            ),
            flush=True,
        )
        # pylint: disable-next=protected-access
        os._exit(0)

    os.waitpid(pid, 0)
    print(
        json.dumps(
            {
                "parent_pid": parent_pid,
                "parent_resource_pid": parent_resource_pid,
                "parent_tracer_pid": parent_tracer_pid,
                "parent_resource_pid_after_fork": tracer_provider.resource.attributes[
                    PROCESS_PID
                ],
                "parent_tracer_pid_after_fork": tracer.resource.attributes[
                    PROCESS_PID
                ],
            }
        ),
        flush=True,
    )


if __name__ == "__main__":
    main()



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=protected-access,unsubscriptable-object

import json
import unittest
from logging import WARNING

from opentelemetry.exporter.otlp.json.common.metrics_encoder import (
    encode_metrics,
)
from opentelemetry.proto_json.collector.metrics.v1.metrics_service import (
    ExportMetricsServiceRequest as JSONExportMetricsServiceRequest,
)
from opentelemetry.sdk.metrics import Exemplar
from opentelemetry.sdk.metrics.export import (
    AggregationTemporality,
    Buckets,
    Metric,
)
from tests import (
    SPAN_ID,
    TIME,
    TRACE_ID,
    assert_proto_json_equal,
    make_exponential_histogram,
    make_gauge,
    make_histogram,
    make_metrics_data,
    make_sum,
)


def _get_first_metric(result):
    return result.resource_metrics[0].scope_metrics[0].metrics[0]


class TestOTLPMetricsEncoder(unittest.TestCase):
    def test_encode_sum(self):
        cases = [
            ("int_value", 33, True),
            ("double_value", 2.98, False),
        ]
        for name, value, is_int in cases:
            with self.subTest(name=name):
                result = encode_metrics(
                    make_metrics_data([make_sum(value=value)])
                )
                encoded = _get_first_metric(result)

                self.assertIsNotNone(encoded.sum)
                self.assertIsNone(encoded.gauge)
                self.assertEqual(
                    encoded.sum.aggregation_temporality,
                    AggregationTemporality.CUMULATIVE,
                )
                self.assertTrue(encoded.sum.is_monotonic)

                dp = encoded.sum.data_points[0]
                self.assertIsNotNone(dp.start_time_unix_nano)
                self.assertIsNotNone(dp.time_unix_nano)
                if is_int:
                    self.assertEqual(dp.as_int, value)
                    self.assertIsNone(dp.as_double)
                else:
                    self.assertEqual(dp.as_double, value)
                    self.assertIsNone(dp.as_int)

    def test_encode_gauge(self):
        cases = [
            ("int_value", 9000, True),
            ("double_value", 52.028, False),
        ]
        for name, value, is_int in cases:
            with self.subTest(name=name):
                result = encode_metrics(
                    make_metrics_data([make_gauge(value=value)])
                )
                encoded = _get_first_metric(result)

                self.assertIsNotNone(encoded.gauge)
                self.assertIsNone(encoded.sum)

                dp = encoded.gauge.data_points[0]
                self.assertIsNotNone(dp.time_unix_nano)
                if is_int:
                    self.assertEqual(dp.as_int, value)
                else:
                    self.assertEqual(dp.as_double, value)

    def test_encode_histogram(self):
        metric = make_histogram(
            exemplars=[
                Exemplar(
                    {"filtered": "banana"}, 298.0, TIME, SPAN_ID, TRACE_ID
                ),
            ],
        )
        result = encode_metrics(make_metrics_data([metric]))
        encoded = _get_first_metric(result)

        self.assertIsNotNone(encoded.histogram)
        self.assertEqual(
            encoded.histogram.aggregation_temporality,
            AggregationTemporality.DELTA,
        )

        dp = encoded.histogram.data_points[0]
        self.assertEqual(dp.count, 5)
        self.assertEqual(dp.sum, 67)
        self.assertEqual(dp.bucket_counts, [1, 4])
        self.assertEqual(dp.explicit_bounds, [10.0, 20.0])
        self.assertEqual(dp.min, 8)
        self.assertEqual(dp.max, 18)
        self.assertEqual(len(dp.exemplars), 1)

    def test_encode_exponential_histogram(self):
        result = encode_metrics(
            make_metrics_data([make_exponential_histogram()])
        )
        encoded = _get_first_metric(result)

        self.assertIsNotNone(encoded.exponential_histogram)
        dp = encoded.exponential_histogram.data_points[0]
        self.assertEqual(dp.count, 10)
        self.assertEqual(dp.sum, 100.5)
        self.assertEqual(dp.scale, 1)
        self.assertEqual(dp.zero_count, 2)
        self.assertIsNotNone(dp.positive)
        self.assertEqual(dp.positive.offset, 0)
        self.assertEqual(dp.positive.bucket_counts, [1, 2, 3])
        self.assertIsNotNone(dp.negative)
        self.assertEqual(dp.negative.offset, 1)
        self.assertEqual(dp.negative.bucket_counts, [1])

    def test_encode_exponential_histogram_empty_buckets(self):
        metric = make_exponential_histogram(
            attributes={},
            count=0,
            sum_value=0.0,
            zero_count=0,
            positive=Buckets(offset=0, bucket_counts=[]),
            negative=Buckets(offset=0, bucket_counts=[]),
            min_value=0.0,
            max_value=0.0,
        )
        result = encode_metrics(make_metrics_data([metric]))
        dp = _get_first_metric(result).exponential_histogram.data_points[0]
        self.assertIsNone(dp.positive)
        self.assertIsNone(dp.negative)

    def test_encode_exemplars(self):
        cases = [
            (
                "float_with_ids",
                Exemplar({"f": "banana"}, 298.0, TIME, SPAN_ID, TRACE_ID),
                True,
                True,
            ),
            (
                "float_without_ids",
                Exemplar({"f": "banana"}, 298.0, TIME, None, None),
                False,
                True,
            ),
            (
                "int_with_ids",
                Exemplar({"f": "banana"}, 42, TIME, SPAN_ID, TRACE_ID),
                True,
                False,
            ),
        ]
        for name, exemplar, has_ids, is_float in cases:
            with self.subTest(name=name):
                metric = make_histogram(
                    attributes={},
                    count=1,
                    sum_value=exemplar.value,
                    bucket_counts=[1],
                    explicit_bounds=[],
                    min_value=exemplar.value,
                    max_value=exemplar.value,
                    exemplars=[exemplar],
                )
                result = encode_metrics(make_metrics_data([metric]))
                enc_ex = (
                    _get_first_metric(result)
                    .histogram.data_points[0]
                    .exemplars[0]
                )
                if has_ids:
                    self.assertTrue(enc_ex.span_id)
                    self.assertTrue(enc_ex.trace_id)
                else:
                    self.assertFalse(enc_ex.span_id)
                    self.assertFalse(enc_ex.trace_id)
                if is_float:
                    self.assertEqual(enc_ex.as_double, exemplar.value)
                else:
                    self.assertEqual(enc_ex.as_int, exemplar.value)

    def test_encode_metrics_resource_and_scope(self):
        result = encode_metrics(
            make_metrics_data(
                [make_gauge(value=1, attributes={})],
                resource_attrs={"service.name": "test-svc"},
                resource_schema_url="res_schema",
                scope_name="my_lib",
                scope_version="2.0",
                scope_schema_url="scope_schema",
            )
        )
        rm = result.resource_metrics[0]
        self.assertEqual(rm.schema_url, "res_schema")
        self.assertEqual(rm.resource.attributes[0].key, "service.name")

        sm = rm.scope_metrics[0]
        self.assertEqual(sm.schema_url, "scope_schema")
        self.assertEqual(sm.scope.name, "my_lib")
        self.assertEqual(sm.scope.version, "2.0")

    def test_encode_metrics_to_dict(self):
        result = encode_metrics(
            make_metrics_data([make_sum(name="sum_int", value=33)])
        )
        result_dict = result.to_dict()

        self.assertIn("resourceMetrics", result_dict)
        rm = result_dict["resourceMetrics"][0]
        self.assertIn("scopeMetrics", rm)

        dp = rm["scopeMetrics"][0]["metrics"][0]["sum"]["dataPoints"][0]
        self.assertIsInstance(dp["startTimeUnixNano"], str)
        self.assertIsInstance(dp["timeUnixNano"], str)
        self.assertIsInstance(dp["asInt"], str)
        self.assertEqual(dp["asInt"], "33")

    def test_encode_metrics_json_roundtrip(self):
        metrics = [
            make_sum(name="s", value=33),
            make_gauge(name="g", value=52.028),
            make_histogram(
                name="h",
                exemplars=[
                    Exemplar({"f": "b"}, 298.0, TIME, SPAN_ID, TRACE_ID),
                ],
            ),
            make_exponential_histogram(name="eh"),
        ]
        result = encode_metrics(make_metrics_data(metrics))
        json_str = result.to_json()
        roundtripped = JSONExportMetricsServiceRequest.from_dict(
            json.loads(json_str)
        )
        assert_proto_json_equal(self, result, roundtripped)

    def test_unsupported_metric_type(self):
        metric = Metric(
            name="unsupported",
            description="foo",
            unit="s",
            data=None,
        )
        with self.assertLogs(level=WARNING):
            result = encode_metrics(make_metrics_data([metric]))
        encoded = _get_first_metric(result)
        self.assertEqual(encoded.name, "unsupported")
        self.assertIsNone(encoded.gauge)
        self.assertIsNone(encoded.sum)
        self.assertIsNone(encoded.histogram)
        self.assertIsNone(encoded.exponential_histogram)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from unittest.mock import patch

from opentelemetry import context
from opentelemetry.context.contextvars_context import ContextVarsRuntimeContext

# pylint: disable=import-error,no-name-in-module
from tests.context.base_context import ContextTestCases


class TestContextVarsContext(ContextTestCases.BaseTest):
    # pylint: disable=invalid-name
    def setUp(self) -> None:
        super().setUp()
        self.mock_runtime = patch.object(
            context,
            "_RUNTIME_CONTEXT",
            ContextVarsRuntimeContext(),
        )
        self.mock_runtime.start()

    # pylint: disable=invalid-name
    def tearDown(self) -> None:
        super().tearDown()
        self.mock_runtime.stop()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: opentelemetry/proto/common/v1/common.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*opentelemetry/proto/common/v1/common.proto\x12\x1dopentelemetry.proto.common.v1\"\xad\x02\n\x08\x41nyValue\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12\x14\n\nbool_value\x18\x02 \x01(\x08H\x00\x12\x13\n\tint_value\x18\x03 \x01(\x03H\x00\x12\x16\n\x0c\x64ouble_value\x18\x04 \x01(\x01H\x00\x12@\n\x0b\x61rray_value\x18\x05 \x01(\x0b\x32).opentelemetry.proto.common.v1.ArrayValueH\x00\x12\x43\n\x0ckvlist_value\x18\x06 \x01(\x0b\x32+.opentelemetry.proto.common.v1.KeyValueListH\x00\x12\x15\n\x0b\x62ytes_value\x18\x07 \x01(\x0cH\x00\x12\x1f\n\x15string_value_strindex\x18\x08 \x01(\x05H\x00\x42\x07\n\x05value\"E\n\nArrayValue\x12\x37\n\x06values\x18\x01 \x03(\x0b\x32\'.opentelemetry.proto.common.v1.AnyValue\"G\n\x0cKeyValueList\x12\x37\n\x06values\x18\x01 \x03(\x0b\x32\'.opentelemetry.proto.common.v1.KeyValue\"e\n\x08KeyValue\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x36\n\x05value\x18\x02 \x01(\x0b\x32\'.opentelemetry.proto.common.v1.AnyValue\x12\x14\n\x0ckey_strindex\x18\x03 \x01(\x05\"\x94\x01\n\x14InstrumentationScope\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12;\n\nattributes\x18\x03 \x03(\x0b\x32\'.opentelemetry.proto.common.v1.KeyValue\x12 \n\x18\x64ropped_attributes_count\x18\x04 \x01(\r\"X\n\tEntityRef\x12\x12\n\nschema_url\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x0f\n\x07id_keys\x18\x03 \x03(\t\x12\x18\n\x10\x64\x65scription_keys\x18\x04 \x03(\tB{\n io.opentelemetry.proto.common.v1B\x0b\x43ommonProtoP\x01Z(go.opentelemetry.io/proto/otlp/common/v1\xaa\x02\x1dOpenTelemetry.Proto.Common.V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'opentelemetry.proto.common.v1.common_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n io.opentelemetry.proto.common.v1B\013CommonProtoP\001Z(go.opentelemetry.io/proto/otlp/common/v1\252\002\035OpenTelemetry.Proto.Common.V1'
  _globals['_ANYVALUE']._serialized_start=78
  _globals['_ANYVALUE']._serialized_end=379
  _globals['_ARRAYVALUE']._serialized_start=381
  _globals['_ARRAYVALUE']._serialized_end=450
  _globals['_KEYVALUELIST']._serialized_start=452
  _globals['_KEYVALUELIST']._serialized_end=523
  _globals['_KEYVALUE']._serialized_start=525
  _globals['_KEYVALUE']._serialized_end=626
  _globals['_INSTRUMENTATIONSCOPE']._serialized_start=629
  _globals['_INSTRUMENTATIONSCOPE']._serialized_end=777
  _globals['_ENTITYREF']._serialized_start=779
  _globals['_ENTITYREF']._serialized_end=867
# @@protoc_insertion_point(module_scope)
